###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Relative package containing all things handling GUI elements."""
# TODO: Redo ExceptionLogger implementation, with more functionality given to ExceptionReporter.

__all__ = (
    'app',
    'AppWindow',
    'ExceptionLogger',
    'ExceptionReporter',
    'ExternalTextBrowser',
    'FileContextMenu',
    'GetterApp',
    'HelpContextMenu',
    'HistoryComboBox',
    'LicenseViewer',
    'PasteLineEdit',
    'ReadmeViewer',
    'ScanSelectorDialog',
    'SettingsWindow',
    'Theme',
    'ToolsContextMenu',
    'tr'
)

from .aliases import app
from .aliases import tr
from .app import GetterApp
from .app import Theme
from .menus import CacheIndexContextMenu
from .menus import ColumnContextMenu
from .menus import FileContextMenu
from .menus import HelpContextMenu
from .menus import ToolsContextMenu
from .widgets import CacheExplorer
from .widgets import ExceptionLogger
from .widgets import ExternalTextBrowser
from .widgets import HistoryComboBox
from .widgets import PasteLineEdit
from .windows import AppWindow
from .windows import ChangelogViewer
from .windows import ExceptionReporter
from .windows import LicenseViewer
from .windows import ReadmeViewer
from .windows import ScanSelectorDialog
from .windows import SettingsWindow



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Package containing GUI elements meant to be used as windows."""

__all__ = (
    'AppWindow',
    'ChangelogViewer',
    'ExceptionReporter',
    'LicenseViewer',
    'ReadmeViewer',
    'ScanSelectorDialog',
    'SettingsWindow',
)

from .application import AppWindow
from .changelog_viewer import ChangelogViewer
from .exception_reporter import ExceptionReporter
from .license_viewer import LicenseViewer
from .readme_viewer import ReadmeViewer
from .scan_selector_dialog import ScanSelectorDialog
from .settings import SettingsWindow



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.
#
# https://github.com/pypa/packaging/blob/main/LICENSE
"""This file has been refurbished by Cubicpath@Github to follow modern typing and documentation practices."""
from __future__ import annotations

__all__ = (
    'BaseVersion',
    'parse',
    'Version',
    'LegacyVersion',
    'InvalidVersion',
    'VERSION_PATTERN'
)

import itertools
import re
from collections import namedtuple
from collections.abc import Callable
from collections.abc import Iterator
from typing import Any
from typing import SupportsInt

from ._structures import Infinity
from ._structures import InfinityType
from ._structures import NegativeInfinity
from ._structures import NegativeInfinityType


InfiniteTypes = InfinityType | NegativeInfinityType
PrePostDevType = InfiniteTypes | tuple[str, int]
SubLocalType = InfiniteTypes | int | str
LocalType = NegativeInfinityType | tuple[
    SubLocalType |
    tuple[SubLocalType, str] |
    tuple[NegativeInfinityType, SubLocalType],
    ...,
],
CmpKey = tuple[
    int, tuple[int, ...], PrePostDevType, PrePostDevType, PrePostDevType, LocalType
]
LegacyCmpKey = tuple[int, tuple[str, ...]]
VersionComparisonMethod = Callable[
    [CmpKey | LegacyCmpKey, CmpKey | LegacyCmpKey], bool
]


_Version = namedtuple(
    '_Version', ['epoch', 'release', 'dev', 'pre', 'post', 'local']
)


def parse(version: str) -> LegacyVersion | Version:
    """Parse the given version string and return either a :class:`Version` object
    or a :class:`LegacyVersion` object depending on if the given version is
    a valid PEP 440 version or a legacy version.
    """
    try:
        return Version(version)
    except InvalidVersion:
        return LegacyVersion(version)


class InvalidVersion(ValueError):
    """An invalid version was found, users should refer to PEP 440."""


class BaseVersion:
    _key: CmpKey | LegacyCmpKey = None

    def __hash__(self) -> int:
        return hash(self._key)

    def __lt__(self, other: BaseVersion) -> bool:
        return self._compare(other, lambda s, o: s < o)

    def __le__(self, other: BaseVersion) -> bool:
        return self._compare(other, lambda s, o: s <= o)

    def __eq__(self, other: Any) -> bool:
        return self._compare(other, lambda s, o: s == o)

    def __ge__(self, other: BaseVersion) -> bool:
        return self._compare(other, lambda s, o: s >= o)

    def __gt__(self, other: BaseVersion) -> bool:
        return self._compare(other, lambda s, o: s > o)

    def __ne__(self, other: Any) -> bool:
        return self._compare(other, lambda s, o: s != o)

    def _compare(self, other: Any, method: VersionComparisonMethod) -> bool | NotImplemented:
        if not isinstance(other, BaseVersion):
            return NotImplemented

        return method(self._key, other._key)


class LegacyVersion(BaseVersion):
    def __init__(self, version: str) -> None:
        self._version = str(version)
        self._key = _legacy_cmpkey(self._version)

    def __str__(self) -> str:
        return self._version

    def __repr__(self) -> str:
        return f'<{type(self).__name__}({repr(str(self))})>'

    @property
    def public(self) -> str:
        return self._version

    @property
    def base_version(self) -> str:
        return self._version

    @property
    def epoch(self) -> int:
        return -1

    @property
    def release(self) -> None:
        return None

    @property
    def pre(self) -> None:
        return None

    @property
    def post(self) -> None:
        return None

    @property
    def dev(self) -> None:
        return None

    @property
    def local(self) -> None:
        return None

    @property
    def is_prerelease(self) -> bool:
        return False

    @property
    def is_postrelease(self) -> bool:
        return False

    @property
    def is_devrelease(self) -> bool:
        return False


_legacy_version_component_re = re.compile(r'(\d+ | [a-z]+ | \.| -)', re.VERBOSE)

_legacy_version_replacement_map = {
    'pre': 'c',
    'preview': 'c',
    '-': 'final-',
    'rc': 'c',
    'dev': '@',
}


def _parse_version_parts(s: str) -> Iterator[str]:
    for part in _legacy_version_component_re.split(s):
        part = _legacy_version_replacement_map.get(part, part)

        if not part or part == '.':
            continue

        if part[:1] in '0123456789':
            # pad for numeric comparison
            yield part.zfill(8)
        else:
            yield '*' + part

    # ensure that alpha/beta/candidate are before final
    yield '*final'


def _legacy_cmpkey(version: str) -> LegacyCmpKey:
    # We hardcode an epoch of -1 here. A PEP 440 version can only have a epoch
    # greater than or equal to 0. This will effectively put the LegacyVersion,
    # which uses the defacto standard originally implemented by setuptools,
    # as before all PEP 440 versions.
    epoch = -1

    # This scheme is taken from pkg_resources.parse_version setuptools prior to
    # it's adoption of the packaging library.
    parts: list[str] = []
    for part in _parse_version_parts(version.lower()):
        if part.startswith('*'):
            # remove "-" before a prerelease tag
            if part < '*final':
                while parts and parts[-1] == '*final-':
                    parts.pop()

            # remove trailing zeros from each series of numeric parts
            while parts and parts[-1] == '00000000':
                parts.pop()

        parts.append(part)

    return epoch, tuple(parts)


# Deliberately not anchored to the start and end of the string, to make it
# easier for 3rd party code to reuse
VERSION_PATTERN = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
"""


class Version(BaseVersion):

    _regex = re.compile(r'^\s*' + VERSION_PATTERN + r'\s*$', re.VERBOSE | re.IGNORECASE)

    def __init__(self, version: str) -> None:
        # Validate the version and parse it into pieces
        match = self._regex.search(version)
        if not match:
            raise InvalidVersion(f'Invalid version: \'{version}\'')

        # Store the parsed out pieces of the version
        self._version = _Version(
            epoch=int(match.group('epoch')) if match.group('epoch') else 0,
            release=tuple(int(i) for i in match.group('release').split('.')),
            pre=_parse_letter_version(match.group('pre_l'), match.group('pre_n')),
            post=_parse_letter_version(
                match.group('post_l'), match.group('post_n1') or match.group('post_n2')
            ),
            dev=_parse_letter_version(match.group('dev_l'), match.group('dev_n')),
            local=_parse_local_version(match.group('local')),
        )

        # Generate a key which will be used for sorting
        self._key = _cmpkey(
            self._version.epoch,
            self._version.release,
            self._version.pre,
            self._version.post,
            self._version.dev,
            self._version.local,
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}({repr(str(self))})>'

    def __str__(self) -> str:
        parts = []

        # Epoch
        if self.epoch != 0:
            parts.append(f'{self.epoch}!')

        # Release segment
        parts.append('.'.join(str(x) for x in self.release))

        # Pre-release
        if self.pre is not None:
            parts.append(''.join(str(x) for x in self.pre))

        # Post-release
        if self.post is not None:
            parts.append(f'.post{self.post}')

        # Development release
        if self.dev is not None:
            parts.append(f'.{self.dev}')

        # Local version segment
        if self.local is not None:
            parts.append(f'+{self.local}')

        return ''.join(parts)

    @property
    def epoch(self) -> int:
        return self._version.epoch

    @property
    def release(self) -> tuple[int, ...]:
        return self._version.release

    @property
    def pre(self) -> tuple[str, int] | None:
        return self._version.pre

    @property
    def post(self) -> tuple[str, int] | None:
        return self._version.post[1] if self._version.post else None

    @property
    def dev(self) -> tuple[str, int] | None:
        return self._version.dev[1] if self._version.dev else None

    @property
    def local(self) -> str | None:
        if self._version.local:
            return '.'.join(str(x) for x in self._version.local)
        else:
            return None

    @property
    def public(self) -> str:
        return str(self).split('+', 1)[0]

    @property
    def base_version(self) -> str:
        parts = []

        # Epoch
        if self.epoch != 0:
            parts.append(f'{self.epoch}!')

        # Release segment
        parts.append('.'.join(str(x) for x in self.release))

        return ''.join(parts)

    @property
    def is_prerelease(self) -> bool:
        return self.dev is not None or self.pre is not None

    @property
    def is_postrelease(self) -> bool:
        return self.post is not None

    @property
    def is_devrelease(self) -> bool:
        return self.dev is not None

    @property
    def major(self) -> int:
        return self.release[0] if len(self.release) >= 1 else 0

    @property
    def minor(self) -> int:
        return self.release[1] if len(self.release) >= 2 else 0

    @property
    def micro(self) -> int:
        return self.release[2] if len(self.release) >= 3 else 0


def _parse_letter_version(letter: str, number: str | bytes | SupportsInt) -> tuple[str, int] | None:
    if letter:
        # We consider there to be an implicit 0 in a pre-release if there is
        # not a numeral associated with it.
        if number is None:
            number = 0

        # We normalize any letters to their lower case form
        letter = letter.lower()

        # We consider some words to be alternate spellings of other words and
        # in those cases we want to normalize the spellings to our preferred
        # spelling.
        if letter == 'alpha':
            letter = 'a'
        elif letter == 'beta':
            letter = 'b'
        elif letter in ['c', 'pre', 'preview']:
            letter = 'rc'
        elif letter in ['rev', 'r']:
            letter = 'post'

        return letter, int(number)
    if not letter and number:
        # We assume if we are given a number, but we are not given a letter
        # then this is using the implicit post release syntax (e.g. 1.0-1)
        letter = 'post'

        return letter, int(number)

    return None


_local_version_separators = re.compile(r'[._-]')


def _parse_local_version(local: str | None) -> LocalType | None:
    """Takes a string like abc.1.twelve and turns it into ("abc", 1, "twelve")."""
    if local is not None:
        return tuple(
            part.lower() if not part.isdigit() else int(part)
            for part in _local_version_separators.split(local)
        )
    return None


def _cmpkey(
    epoch: int,
    release: tuple[int, ...],
    pre: tuple[str, int] | None,
    post: tuple[str, int] | None,
    dev: tuple[str, int] | None,
    local: tuple[SubLocalType] | None,
) -> CmpKey:
    # When we compare a release version, we want to compare it with all of the
    # trailing zeros removed. So we'll use a reverse the list, drop all the now
    # leading zeros until we come to something non zero, then take the rest
    # re-reverse it back into the correct order and make it a tuple and use
    # that for our sorting key.
    _release = tuple(
        reversed(list(itertools.dropwhile(lambda x: x == 0, reversed(release))))
    )

    # We need to "trick" the sorting algorithm to put 1.0.dev0 before 1.0a0.
    # We'll do this by abusing the pre segment, but we _only_ want to do this
    # if there is not a pre or a post segment. If we have one of those then
    # the normal sorting rules will handle this case correctly.
    if pre is None and post is None and dev is not None:
        _pre: PrePostDevType = NegativeInfinity
    # Versions without a pre-release (except as noted above) should sort after
    # those with one.
    elif pre is None:
        _pre = Infinity
    else:
        _pre = pre

    # Versions without a post segment should sort before those with one.
    if post is None:
        _post: PrePostDevType = NegativeInfinity

    else:
        _post = post

    # Versions without a development segment should sort after those with one.
    if dev is None:
        _dev: PrePostDevType = Infinity

    else:
        _dev = dev

    if local is None:
        # Versions without a local segment should sort before those with one.
        _local: LocalType = NegativeInfinity
    else:
        # Versions with a local segment need that segment parsed to implement
        # the sorting rules in PEP440.
        # - Alpha numeric segments sort before numeric segments
        # - Alpha numeric segments sort lexicographically
        # - Numeric segments sort numerically
        # - Shorter versions sort before longer versions when the prefixes
        #   match exactly
        _local = tuple(
            (i, '') if isinstance(i, int) else (NegativeInfinity, i) for i in local
        )

    return epoch, _release, _pre, _post, _dev, _local



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""CacheExplorer implementation."""
from __future__ import annotations

__all__ = (
    'CacheExplorer',
)

from pathlib import Path

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...constants import *
from ...events import EventBus
from ...models import DeferredCallable
from ...tomlfile import TomlEvents
from ...utils import init_objects
from ..aliases import app


class _IconProvider(QAbstractFileIconProvider):
    NULL_ICON = QIcon()

    def __init__(self, cache_explorer: CacheExplorer):
        super().__init__()
        self._cache_explorer = cache_explorer
        self._icon_mode: int = app().settings['gui/cache_explorer/icon_mode']  # type: ignore
        self._fallback_provider = QFileIconProvider()

        EventBus['settings'].subscribe(
            self.on_mode_change, TomlEvents.Set,
            event_predicate=lambda event: event.new != event.old and event.key == 'gui/cache_explorer/icon_mode'
        )

    def icon(self, info: QFileInfo | QAbstractFileIconProvider.IconType) -> QIcon:
        """Get the icon for a path using the given information.

        settings['gui/cache_explorer/icon_mode'] changes how this function works.
        0 --- No Icons
        1 --- Default Icons
        2 --- Preview Images

        :param info: File info to associate icon with.
        :return: Icon read from image, or icon from the fallback QFileIconProvider.
        """
        if not self._icon_mode:
            return _IconProvider.NULL_ICON

        if isinstance(info, QFileInfo):
            if info.isDir() and info.fileName() == 'cached_requests':
                return self._cache_explorer.style().standardIcon(QStyle.StandardPixmap.SP_DialogSaveButton)

            # Preview images
            if self._icon_mode == 2:
                if info.isFile() and info.suffix().lstrip('.') in SUPPORTED_IMAGE_EXTENSIONS:
                    icon = QIcon()
                    icon.addFile(info.filePath(), QSize(24, 24))
                    return icon

        return self._fallback_provider.icon(info)

    def on_mode_change(self, event: TomlEvents.Set):
        """Set the icon mode to the one selected in settings.

        :raises TypeError: If new value from event is not an int.
        """
        value = event.new
        if not isinstance(value, int):
            raise TypeError(f'Event value "{value}" is not of type {int}.')

        self._icon_mode = value
        self._cache_explorer.model().setIconProvider(self)


class CacheExplorer(QTreeView):
    """:py:class:`QTreeView` used to interact with the app's cache directory.

    Has a native :py:class:`QFileSystemModel` and item context menu.
    """

    openFileInView = Signal(str)

    def __init__(self, parent, *args, **kwargs) -> None:
        """Create a new :py:class:`CacheExplorer` widget for viewing cached requests and files."""
        super().__init__(parent, *args, **kwargs)

        init_objects({
            (model := QFileSystemModel(self)): {
                'readOnly': True,
                'rootPath': HI_CACHE_PATH.absolute().as_posix(),
                'iconProvider': _IconProvider(self),
                'nameFilters': [f'*.{ext}' for ext in SUPPORTED_IMAGE_EXTENSIONS],
                'nameFilterDisables': True
            },

            self: {
                'model': model,
                'rootIndex': model.index(HI_CACHE_PATH.absolute().as_posix()),
                'indentation': 12,
                'contextMenuPolicy': Qt.ContextMenuPolicy.CustomContextMenu,
                'customContextMenuRequested': self.on_custom_context_menu,
                'doubleClicked': self.on_double_click
            },

            self.header(): {
                'size': {'fixed': (None, 26)},
                'sectionsClickable': True,
                'contextMenuPolicy': Qt.ContextMenuPolicy.CustomContextMenu,
                'customContextMenuRequested': self.on_header_custom_context_menu
            }
        })

        # Hide all but Name by default
        for index in (1, 2, 3):  # SIZE, FILE TYPE, DATE MODIFIED
            self.hideColumn(index)

    def delete_index(self, index: QModelIndex) -> None:
        """Delete a given index from the file system after prompting the user.

        If the path resolves to a directory, all items in the directory are recursively removed.

        :param index: The index to get the filepath from and delete.
        """
        file_path: Path = Path(self.model().filePath(index))
        path_type: str = 'Folder' if file_path.is_dir() else 'File'

        consent: bool = app().show_dialog(
            'warnings.delete_path', self, (
                QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel
            ), default_button=QMessageBox.StandardButton.Cancel,
            title_args=(path_type,),
            description_args=(path_type, file_path)
        ).role == QMessageBox.ButtonRole.AcceptRole

        if consent:
            if file_path.is_dir():
                QDir(file_path).removeRecursively()
            else:
                file_path.unlink()

    def expandAll(self) -> None:
        """Call super().expandall ~50 times."""
        timer = QTimer(self)
        timer.timeout.connect(              # pyright: ignore[reportGeneralTypeIssues]
            DeferredCallable(super().expandAll)
        )
        timer.start(5)

        timer2 = QTimer(self)
        timer2.setSingleShot(True)
        timer2.timeout.connect(timer.stop)  # pyright: ignore[reportGeneralTypeIssues]
        timer2.start(500)

    def expandRecursively(self, *args, **kwargs) -> None:
        """Call super().expandRecursively ~50 times."""
        timer = QTimer(self)
        timer.timeout.connect(              # pyright: ignore[reportGeneralTypeIssues]
            DeferredCallable(super().expandRecursively, *args, **kwargs)
        )
        timer.start(5)

        timer2 = QTimer(self)
        timer2.setSingleShot(True)
        timer2.timeout.connect(timer.stop)  # pyright: ignore[reportGeneralTypeIssues]
        timer2.start(500)

    def model(self) -> QFileSystemModel:
        """:return: The file system model for the CacheExplorer."""
        # pylint: disable=useless-super-delegation
        return super().model()  # type: ignore

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """Map keys to actions."""
        super().keyPressEvent(event)

        match event.key():
            case Qt.Key.Key_Delete:
                # Delete Path at Selected Index
                for i in self.selectedIndexes():
                    self.delete_index(i)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """Resize "Name" column.py on click/expansion."""
        super().mousePressEvent(event)
        self.resizeColumnToContents(0)

    def on_header_custom_context_menu(self, point: QPoint) -> None:
        """Ran when the customContextMenuRequested signal is emitted.

        :param point: The point to place the context menu.
        """
        from ..menus import ColumnContextMenu

        index = self.indexAt(point)

        if index.isValid():
            menu = ColumnContextMenu(self, {0})
            menu.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

            menu.exec(self.viewport().mapToGlobal(point))

    def on_custom_context_menu(self, point: QPoint) -> None:
        """Ran when the customContextMenuRequested signal is emitted.

        :param point: The point to place the context menu.
        """
        from ..menus import CacheIndexContextMenu

        index = self.indexAt(point)

        if index.isValid():
            menu = CacheIndexContextMenu(self, index)
            menu.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

            menu.exec(self.viewport().mapToGlobal(point))

    def on_double_click(self, index: QModelIndex) -> None:
        """Ran when the doubleClicked signal is emitted.

        :param index: The model index which was double-clicked.
        """
        if index.isValid():
            self.openFileInView.emit(self.model().filePath(index))



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Version checking and validation."""
from __future__ import annotations

__all__ = (
    'get_version',
    'is_greater_version',
    'VersionChecker',
)

from datetime import datetime
from importlib.metadata import version

from PySide6.QtCore import *

from .._vendor.packaging import BaseVersion
from .._vendor.packaging import LegacyVersion
from .._vendor.packaging import parse_version
from .._vendor.packaging import Version
from .._version import __version__
from ..constants import *
from ..utils.package import has_package
from .manager import NetworkSession
from .manager import Response


def get_version(package_name: str) -> LegacyVersion | Version | None:
    """Return the :py:class:`Version` of the given package if it is installed. Else return None."""
    if has_package(package_name):
        return parse_version(version(package_name))


def is_greater_version(version1: BaseVersion | str, version2: BaseVersion | str) -> bool:
    """Return whether ``version1`` is greater than ``version2``."""
    if not isinstance(version1, BaseVersion):
        version1 = parse_version(version1)
    if not isinstance(version2, BaseVersion):
        version2 = parse_version(version2)

    return version1 > version2


class VersionChecker(QObject):
    """Checks for the latest versions of packages."""

    checked = Signal(str)
    newerVersion = Signal(str, str)

    def __init__(self, parent: QObject | None = None) -> None:
        """Create a new :py:class:`VersionChecker` and initialize its :py:class:`NetworkSession`."""
        super().__init__(parent)

        self.session: NetworkSession = NetworkSession(self)

    def check_version(self, package_name: str = HI_PACKAGE_NAME) -> None:
        """Check the latest version of the given package on PyPI.

        If the latest version is greater than the current version,
        emit the ``newerVersion`` signal with the given name and the latest version as arguments.

        Emit the ``checked`` signal after a successful call.

        :param package_name: The package to look up on PyPI.
        """
        def handle_reply(reply: Response):
            # Sort versions on date released
            versions: list[str] = sorted(
                releases := reply.json['releases'],
                key=lambda v: datetime.strptime(releases[v][0]['upload_time_iso_8601'], '%Y-%m-%dT%H:%M:%S.%fZ')
            )

            # Get local version of given package. Use __version__ attribute for own package.
            local_version: BaseVersion | str
            if package_name != HI_PACKAGE_NAME:
                if (ver := get_version(package_name)) is None:
                    return
                local_version = ver
            else:
                local_version = __version__

            # Get the latest version and compare to current version. Emit newerVersion if greater.
            latest: str = versions[-1]
            if is_greater_version(latest, local_version):
                self.newerVersion.emit(package_name, latest)
            self.checked.emit(package_name)

        self.session.get(f'https://pypi.org/pypi/{package_name.replace("_", "-").strip()}/json', finished=handle_reply)



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""ExceptionReporter implementation."""
from __future__ import annotations

__all__ = (
    'ExceptionReporter',
)

from getpass import getuser

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ..._version import __version__
from ...events import EventBus
from ...exception_hook import ExceptionEvent
from ...models import DeferredCallable
from ...models import DistributedCallable
from ...utils import delete_layout_widgets
from ...utils import encode_url_params
from ...utils import format_tb
from ...utils import init_layouts
from ...utils import init_objects
from ..aliases import app
from ..aliases import tr
from ..widgets import ExceptionLogger
from ..widgets import ExternalTextBrowser


class ExceptionReporter(QWidget):
    """A :py:class:`QWidget` that displays logged exceptions and their traceback."""

    def __init__(self, logger: ExceptionLogger) -> None:
        """Create a new :py:class:`ExceptionReporter` and initialize UI."""
        super().__init__()
        EventBus['exceptions'].subscribe(DeferredCallable(self.reload_exceptions), ExceptionEvent)

        self.selected: int = 0
        self.logger: ExceptionLogger = logger
        self.setWindowTitle(tr('gui.exception_reporter.title'))
        self.setWindowIcon(self.logger.icon())
        self.resize(QSize(750, 400))
        self._init_ui()

    def _init_ui(self) -> None:
        self.left_panel = QFrame(self)
        self.right_panel = QFrame(self)

        self.trace_back_viewer = ExternalTextBrowser(self)
        self.clear_button = QPushButton(self.right_panel)
        self.report_button = QPushButton(self.right_panel)

        # Define the scroll area/widget for the left panel
        self.scroll_area = QScrollArea(self)
        self.scroll_widget = QWidget(self.scroll_area)
        self.scroll_widget.setLayout(QVBoxLayout())
        self.scroll_widget.layout().setAlignment(Qt.AlignmentFlag.AlignTop)

        init_objects({
            self.scroll_area: {
                'widget': self.scroll_widget,
                'widgetResizable': True
            },

            self.left_panel: {'layout': QVBoxLayout()},
            self.right_panel: {'layout': QVBoxLayout()},
            (left_label := QLabel(self.left_panel)): {},
            (right_label := QLabel(self.right_panel)): {},

            (clear_all_button := QPushButton(self.left_panel)): {
                'clicked': self.clear_all_exceptions
            },

            self.trace_back_viewer: {
                'disabled': True,
                'font': QFont('consolas', 10),
                'lineWrapMode': QTextEdit.LineWrapMode.NoWrap
            },

            self.clear_button: {
                'disabled': True,
                'size': {'maximum': (self.report_button.width() // 1, None)},
                'clicked': self.clear_current_exception
            },

            self.report_button: {
                'disabled': True,
                'clicked': self.report_current_exception
            }
        })

        app().init_translations({
            self.setWindowTitle: 'gui.exception_reporter.title',

            # Labels
            left_label.setText: 'gui.exception_reporter.exception_list',
            right_label.setText: 'gui.exception_reporter.traceback_label',

            # Buttons
            clear_all_button.setText: 'gui.exception_reporter.clear_all',
            self.clear_button.setText: 'gui.exception_reporter.clear',
            self.report_button.setText: 'gui.exception_reporter.report'
        })

        init_layouts({
            # Left frame layout
            self.left_panel.layout(): {
                'items': [left_label, self.scroll_area, clear_all_button]
            },

            # Right frame layouts
            (buttons := QHBoxLayout()): {
                'items': [self.report_button, self.clear_button]
            },
            self.right_panel.layout(): {
                'items': [right_label, self.trace_back_viewer, buttons]
            },

            # Main layout
            QHBoxLayout(self): {
                'items': [self.left_panel, self.right_panel]
            }
        })

        self.left_panel.setLayout(QVBoxLayout())
        self.right_panel.setLayout(QVBoxLayout())

    def report_current_exception(self) -> None:
        """Report the current exception to the exception logger."""
        selected = self.logger.exception_log[self.selected]

        exc_name: str = type(selected.exception).__name__
        exc_msg: str = str(selected.exception).rstrip('.')
        exc_tb: str = format_tb(selected.traceback).strip().replace(getuser(), '%USERNAME%')

        base: str = 'https://github.com/Cubicpath/HaloInfiniteGetter/issues/new'
        params: dict[str, str] = {
            'template': 'bug-report.yaml',
            'assignees': 'Cubicpath',
            'labels': 'bug',
            'projects': 'Cubicpath/HaloInfiniteGetter/2',
            'title': f'[Bug]: ({exc_name}) {exc_msg}',
            'version': f'v{__version__}',
            'logs': f'{exc_name}: {exc_msg}\n\n{exc_tb}'
        }
        QDesktopServices.openUrl(QUrl(f'{base}?{encode_url_params(params)}'))

    def disable_exception_widgets(self) -> None:
        """Disables the widgets on the right pane."""
        self.trace_back_viewer.clear()
        self.trace_back_viewer.setDisabled(True)
        self.report_button.setDisabled(True)
        self.clear_button.setDisabled(True)

    def clear_all_exceptions(self) -> None:
        """Clear all exceptions from the list."""
        self.logger.clear_exceptions()
        self.disable_exception_widgets()
        delete_layout_widgets(self.scroll_widget.layout())

    def clear_current_exception(self) -> None:
        """Clear the currently selected exception and removes it from the log."""
        if (item := self.scroll_widget.layout().takeAt(self.selected)) is not None:
            self.logger.remove_exception(self.selected)
            item.widget().deleteLater()
        self.reload_exceptions()
        self.disable_exception_widgets()

    def reload_exceptions(self) -> None:
        """Load the exceptions from the logger."""
        delete_layout_widgets(self.scroll_widget.layout())
        for i, error in enumerate(self.logger.exception_log):
            button = QPushButton(f'{type(error.exception).__name__}: {error.exception}', self.scroll_widget)
            clicked = button.clicked  # pyright: ignore[reportGeneralTypeIssues]
            clicked.connect(DeferredCallable(setattr, self, 'selected', i))
            clicked.connect(DistributedCallable((
                self.clear_button.setDisabled,
                self.report_button.setDisabled,
                self.trace_back_viewer.setDisabled
            ), False))
            clicked.connect(DeferredCallable(self.trace_back_viewer.setText, DeferredCallable(
                tr, 'gui.exception_reporter.traceback_view',
                type(error.exception).__name__, error.exception, format_tb(error.traceback),
                key_eval=False
            )))

            button.setStyleSheet('text-align:left;')
            button.setMaximumWidth((self.size().width() // 3) - 50)

            self.scroll_widget.layout().addWidget(button)

    def resizeEvent(self, event: QResizeEvent) -> None:
        """Resizes the left panel to better fit the window."""
        super().resizeEvent(event)
        self.left_panel.setMaximumWidth(event.size().width() // 3)
        for i in range(self.scroll_widget.layout().count()):
            self.scroll_widget.layout().itemAt(i).widget().setMaximumWidth((event.size().width() // 3) - 50)
        event.accept()



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Package-related utility functions."""
from __future__ import annotations

__all__ = (
    'current_requirement_licenses',
    'current_requirement_names',
    'current_requirement_versions',
    'has_package',
)


def current_requirement_licenses(
        package: str, recursive: bool = False, include_extras: bool = False
) -> dict[str, list[tuple[str, str]]]:
    """Return the current licenses for the requirements of the given package.

    :param package: Package name to search.
    :param recursive: Whether to get the requirements of the requirements of the given package.
    :param include_extras: Whether to include packages installed with extras.
    :return: dict mapping a package name to a list of license title and content pairs.
    """
    from importlib.metadata import metadata
    from pathlib import Path
    from pkg_resources import Distribution
    from pkg_resources import DistributionNotFound
    from pkg_resources import get_distribution

    result: dict[str, list[tuple[str, str]]] = {}
    for requirement in ([package] + current_requirement_names(package, recursive, include_extras)):
        try:
            dist: Distribution = get_distribution(requirement)
        except DistributionNotFound:
            continue

        name: str = dist.project_name.replace('-', '_')
        licenses: list[tuple[str, str]] = []

        # Find the distribution's information directory
        info_path = Path(dist.location) / f'{name}-{dist.version}.dist-info'
        if not info_path.is_dir():
            if (egg_path := info_path.with_name(f'{name}.egg-info')).is_dir():
                info_path = egg_path

        # Find the license file(s)
        license_files: list[Path] = [
            item for item in info_path.iterdir() if (
                item.is_file() and any(keyword in item.name.lower() for keyword in
                                       ('license', 'licence', 'copying', 'copyright', 'notice', 'author'))
            )
        ]

        # If the distribution has no 'License' field, get the distribution's Trove classifier
        if not (dist_license := metadata(name).get('License')):  # type: ignore
            for classifier in metadata(name).get_all('Classifier'):
                if classifier.startswith('License'):
                    # Ex: 'License :: OSI Approved :: MIT License'
                    dist_license = classifier.split('::')[-1].strip()
                    break

        # If there are multiple license/notice files, the title contains the filename
        for file in license_files:
            license_name: str = dist_license if len(license_files) == 1 else f'{dist_license} - {file.name}'
            licenses.append((license_name, file.read_text(encoding='utf8')))

        result[name] = licenses

    return result


def current_requirement_names(package: str, recursive: bool = False, include_extras: bool = False) -> list[str]:
    """Return the current requirement names for the given package.

    :param package: Package name to search.
    :param recursive: Whether to get the requirements of the requirements of the given package.
    :param include_extras: Whether to include packages installed with extras.
    :return: list of package names.
    """
    from importlib.metadata import PackageNotFoundError
    from importlib.metadata import requires

    req_names: set[str] = set()
    try:
        # Get the requirements for the given package, translating None to an empty tuple to allow iteration
        requirements: list[str] = requires(package) or []
    except PackageNotFoundError:
        # If the package is not installed, return an empty list
        return []

    for requirement in requirements:
        if not include_extras and '; extra' in requirement:
            continue

        # Don't include testing extras
        testing_extras = {'dev', 'develop', 'development', 'test', 'testing'}
        if include_extras and requirement.split('extra ==')[-1].strip().strip('"') in testing_extras:
            continue

        # Get the package name from the requirement
        # If no version identifier, assume the requirement string is the package name
        name: str = requirement
        has_version: bool = False
        split_pos: int = 0
        for split_pos, char in enumerate(requirement):
            if not char.isalnum() and char not in ('-', '_'):
                has_version = True
                break

        if has_version:
            name = requirement[:split_pos]

        req_names.add(name)
        if recursive:
            req_names.update(current_requirement_names(name, recursive, include_extras))

    return sorted(req_names, key=str.lower)


def current_requirement_versions(package: str, recursive: bool = False, include_extras: bool = False) -> dict[str, str]:
    """Return the current versions of the installed requirements for the given package.

    :param package: Package name to search.
    :param recursive: Whether to get the requirements of the requirements of the given package.
    :param include_extras: Whether to include packages installed with extras.
    :return: dict mapping package names to their version string.
    """
    from importlib.metadata import version

    return {name: version(name) for
            name in current_requirement_names(package, recursive, include_extras)
            if has_package(name)}


def has_package(package: str) -> bool:
    """Check if the given package is available.

    :param package: Package name to search; hyphen-insensitive.
    :return: Whether the given package name is installed to the current environment.
    """
    from pkg_resources import WorkingSet
    package = package.replace('-', '_')

    return any(
        package == pkg.project_name.replace('-', '_')
        for pkg in WorkingSet()
    )



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Initialize values and runs the application. :py:func:`main` acts as an entry-point."""
from __future__ import annotations

__all__ = (
    'main',
)

import sys

from ._version import __version__
from .exception_hook import ExceptionHook
from .gui import GetterApp
from .utils import patch_windows_taskbar_icon


def main(*args: str) -> int:
    """Run the program. GUI script entrypoint.

    Args are passed to a QApplication instance.
    """
    patch_windows_taskbar_icon(f'cubicpath.{__package__}.app.{__version__}')

    # ExceptionHook is required for subscribing to ExceptionEvents
    with ExceptionHook():
        GetterApp.create(*args)

        app: GetterApp = GetterApp.instance()
        app.windows['app'].show()
        return app.exec()


if __name__ == '__main__':
    main(*sys.argv)



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Defines the HaloInfiniteGetter HTTP Client."""
from __future__ import annotations

__all__ = (
    'Client',
)

import json
import os
from collections.abc import Callable
from pathlib import Path
from typing import Any
from typing import Final

from PySide6.QtCore import *

from ..constants import *
from ..models import CaseInsensitiveDict
from ..utils import dump_data
from ..utils import decode_url
from ..utils import guess_json_utf
from ..utils import hide_windows_file
from ..utils import unique_values
from .manager import NetworkSession
from .manager import Response

_ENDPOINT_PATH: Final[Path] = HI_RESOURCE_PATH / 'wp_endpoint_hosts.json'


class Client(QObject):
    """Asynchronous HTTP REST Client that interfaces with Halo Waypoint to get data."""

    finishedSearch = Signal()
    startedSearch = Signal()

    receivedData = Signal(str, bytes)
    receivedError = Signal(str, int)
    receivedJson = Signal(str, dict)

    def __init__(self, parent: QObject, **kwargs) -> None:
        """Initialize Halo Waypoint Client.

        Order of precedence for token and wpauth values::

            1. token/wpauth kwarg value
            2. HI_SPARTAN_AUTH/HI_WAYPOINT_AUTH environment variables
            3. user's .token/.wpauth config file (Created by application)

        :keyword token: Token to authenticate self to 343 API.
        :keyword wpauth: Halo Waypoint authentication key, allows for creation of 343 auth tokens.
        """
        super().__init__(parent)
        self.endpoints: dict[str, Any] = json.loads(_ENDPOINT_PATH.read_text(encoding='utf8'))
        self._emit_received_signals: bool = True
        self._recursive_calls_in_progress: int = 0

        self.parent_path: str = '/hi/'
        self.sub_host: str = self.endpoints['endpoints']['gameCmsService']['subdomain']
        self.searched_paths: set[str] = set()
        self.finishedSearch.connect(self._on_finished_search)

        self._token: str | None = kwargs.pop('token', os.getenv('HI_SPARTAN_AUTH', None))
        if self._token is None and HI_TOKEN_PATH.is_file():
            self._token = HI_TOKEN_PATH.read_text(encoding='utf8').strip()

        self._wpauth: str | None = kwargs.pop('wpauth', os.getenv('HI_WAYPOINT_AUTH', None))
        if self._wpauth is None and HI_WPAUTH_PATH.is_file():
            self._wpauth = HI_WPAUTH_PATH.read_text(encoding='utf8').strip()

        self.api_session: NetworkSession = NetworkSession(self)
        self.api_session.headers = CaseInsensitiveDict({
            'Accept': ', '.join(('application/json', 'text/plain', '*/*')),
            # 'Accept-Encoding': ', '.join(('gzip', 'deflate', 'br')),
            'Accept-Language': 'en-US',
            'Connection': 'keep-alive',
            'Cache-Control': 'no-cache',
            'DNT': '1',
            'Host': self.host,
            'Origin': 'https://www.halowaypoint.com',
            'Pragma': 'no-cache',
            'Referer': 'https://www.halowaypoint.com/',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'same-site',
            'Sec-GPC': '1',
            'TE': 'trailers',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:100.0) Gecko/20100101 Firefox/100.0',
        })
        self.web_session: NetworkSession = NetworkSession(self)
        self.web_session.headers = self.api_session.headers | {
            'Accept': ','.join(('text/html', 'application/xhtml+xml', 'application/xml;q=0.9',
                                'image/avif', 'image/webp', '*/*;q=0.8')),
            'Host': self.endpoints['webHost'],
            'Sec-Fetch-Dest': 'document',
            'Sec-Fetch-Mode': 'navigate',
            'Sec-Fetch-Site': 'none',
            'Sec-Fetch-User': '?1',
        }

        if self.wpauth:
            self.set_cookie('wpauth', self.wpauth)
            if not self.token:
                self.refresh_auth()

        if self.token:
            self.api_session.headers['x-343-authorization-spartan'] = self.token
            self.set_cookie('343-spartan-token', self.token)

    def _get(self,
             path: str,
             update_auth_on_401: bool = True,
             finished: Callable[[Response], None] | None = None,
             **kwargs
             ) -> None:
        """Get a :py:class:`Response` from HaloWaypoint.

        :param path: path to append to the API root.
        :param update_auth_on_401: run self._refresh_auth if response status code is 401 Unauthorized.
        :param finished: Callback to send finished request to.
        """

        def handle_reply(response: Response):
            if not response.code or response.headers.get('Content-Type') is not None and not response.data:
                print(f'COULDNT GET {response.url}')
                return

            if not response.ok:
                # Handle errors
                if response.code == 401 and update_auth_on_401 and self.wpauth is not None:
                    self.refresh_auth()
                    self._get(path, False, finished, **kwargs)
                elif finished is not None:
                    # Send ERR response to the given consumer
                    finished(response)

            elif finished is not None:
                # Send OK response to the given consumer
                finished(response)

        self.api_session.get(self.api_root + path.strip(), finished=handle_reply, **kwargs)

    def _handle_json(self, data: dict[str, Any], recursive: bool = False):
        """Look through JSON data for resource paths.

        :param data: JSON data to search through for paths.
        :param recursive: Whether to start recursively searching.
        """
        for value in unique_values(data):
            if isinstance(value, str) and (match := HI_PATH_PATTERN.match(value)):
                new_path: str = self.normalize_search_path(match[0])

                if not recursive:
                    self.get_hi_data(new_path, emit_signals=self._emit_received_signals)
                    continue

                # If it's an image, download it then ignore the result
                if match['file_name'].split('.')[-1].lower() in SUPPORTED_IMAGE_EXTENSIONS:
                    if new_path in self.searched_paths:
                        continue

                    self.searched_paths.add(new_path)
                    self.get_hi_data(new_path, emit_signals=self._emit_received_signals)

                # Otherwise, start the process over again
                else:
                    if new_path in self.searched_paths:
                        continue

                    self.searched_paths.add(new_path)
                    self.recursive_search(new_path)

    def _increment_counter(self):
        self._recursive_calls_in_progress += 1
        if self._recursive_calls_in_progress == 1:
            self.startedSearch.emit()

    def _decrement_counter(self):
        self._recursive_calls_in_progress -= 1
        if not self._recursive_calls_in_progress:
            self.finishedSearch.emit()

    def _on_finished_search(self):
        self._emit_received_signals = True
        self.searched_paths.clear()

    def get_hi_data(self,
                    path: str,
                    consumer: Callable[[str, dict[str, Any] | bytes | int], None] | None = None,
                    emit_signals: bool = True
                    ) -> None:
        """Get a resource hosted on the HaloWaypoint API.

        Resulting resource is returned via a positional argument in a `received-` signal or the given consumer.

        :param path: Path to get data from.
        :param consumer: Consumer to send received data to.
        :param emit_signals: Whether to emit received* signals.
        :raises ValueError: If the response is not JSON or a supported image type.
        """
        path = self.normalize_search_path(path)
        os_path: Path = self.to_os_path(path, parent=HI_WEB_DUMP_PATH)

        if not os_path.is_file():
            def handle_reply(response: Response):
                if response.code and not response.ok:
                    print(f'ERROR [{response.code}] for {path} ')
                    if emit_signals:
                        self.receivedError.emit(path, response.code)
                    if consumer is not None:
                        consumer(path, response.code)
                    return

                content_type: str | None = response.headers.get('Content-Type')
                response_data: dict[str, Any] | bytes

                if not content_type:
                    raise ValueError('Successful status code but no Content-Type header.')

                if 'json' in content_type:
                    response_data = response.json
                    if emit_signals:
                        self.receivedJson.emit(path, response_data)
                elif content_type in SUPPORTED_IMAGE_MIME_TYPES:
                    response_data = response.data
                    if emit_signals:
                        self.receivedData.emit(path, response_data)
                else:
                    raise ValueError(f'Unsupported content type received: {content_type}')

                print(f'DOWNLOADED {path} >>> {content_type}')
                dump_data(os_path, response_data)
                if consumer is not None:
                    consumer(path, response_data)

            self._get(path, finished=handle_reply, timeout=120)

        else:
            print(f'READING {path}')
            data: dict[str, Any] | bytes = os_path.read_bytes()
            if os_path.suffix.lstrip('.') in SUPPORTED_IMAGE_EXTENSIONS:
                self.receivedData.emit(path, data)
            else:
                # Assume json if not an image
                data = json.loads(data.decode(encoding=guess_json_utf(data) or 'utf8'))
                self.receivedJson.emit(path, data)

            if consumer is not None:
                consumer(path, data)

    def recursive_search(self, path: str) -> None:
        """Get a file and recursively look through its contents for paths.

        :param path: Path to search.
        """
        self._emit_received_signals = False
        self._increment_counter()
        self.get_hi_data(path, consumer=self.handle_recursive_data, emit_signals=self._emit_received_signals)

    def start_handle_json(self, data: dict[str, Any], recursive: bool = False):
        """Look through JSON data for resource paths.

        Increments and decrements counter to correctly emit the ``finishedSearch`` signal.

        :param data: JSON data to search through for paths.
        :param recursive: Whether to start recursively searching.
        """
        if recursive:
            self._increment_counter()

        self._handle_json(data, recursive)

        if recursive:
            self._decrement_counter()

    def handle_recursive_data(self, path: str, data: dict[str, Any] | bytes | int) -> None:
        """Recursively look through JSON data for resource paths, decrementing counter when finished.

        :param path: Path data is from.
        :param data: Data received from path. If not JSON, return early.
        """
        # This set is shared between all recursive calls, so no duplicate searches should occur
        self.searched_paths.add(path)

        if isinstance(data, (bytes, int)):
            # Return early, decrementing counter
            self._decrement_counter()
            return

        # Iterate over all values in the JSON data
        # This process ignores already-searched values
        self._handle_json(data, recursive=True)
        self._decrement_counter()

    def hidden_key(self) -> str:
        """:return: The first and last 3 characters of the waypoint token, seperated by periods."""
        key = self.wpauth
        if key is not None and len(key) > 6:
            return f'{key[:3]}{"." * 50}{key[-3:]}'
        return 'None'

    def normalize_search_path(self, path: str) -> str:
        """Normalize and expand search paths.

        If path starts with the ``api_root``, assume the contents are already expanded.
        If not explicitly an images file but has an image extension, assume it's from images/file/ directory.
        If not explicitly a progression file but has a json extension, assume it's from progression/file/ directory.

        :param path: Search path to normalize.
        :return: Normalized version of ``path``.
        """
        # Ensure lowercase
        path = path.lower().strip().lstrip('/')
        parent_path = self.parent_path.lower().lstrip('/')
        file_ext = path.split('.')[-1]

        # Expand paths
        if path.startswith(parent_path):
            path = path.removeprefix(parent_path)
        elif not path.startswith('images/file/') and (file_ext in SUPPORTED_IMAGE_EXTENSIONS):
            path = f'images/file/{path}'
        elif not path.startswith('progression/file/') and (file_ext == 'json'):
            path = f'progression/file/{path}'

        return path

    def to_os_path(self, path: str, parent: Path = HI_WEB_DUMP_PATH) -> Path:
        """Translate a given GET path to the equivalent cache location.

        ``path`` is assumed to be normalized.

        :param path: path to translate.
        :param parent: OS path to use as parent directory.
        """
        return parent / self.sub_host.replace('-', '_') / self.parent_path.strip('/') / path

    def to_get_path(self, path: str) -> str:
        """Translate a given cache location to the equivalent GET path."""
        resource = path.split(self.parent_path, maxsplit=1)[-1]
        return resource

    def refresh_auth(self) -> None:
        """Refresh authentication to Halo Waypoint servers.

        wpauth MUST have a value for this to work. A lone 343 spartan token is not enough to generate a new one.
        """
        def handle_reply(_: Response):
            wpauth: str = decode_url(self.web_session.cookies.get('wpauth') or '')
            token: str = decode_url(self.web_session.cookies.get('343-spartan-token') or '')

            if wpauth and self.wpauth != wpauth:
                self.wpauth = wpauth
            if token:
                self.token = token

        self.web_session.get('https://www.halowaypoint.com/', finished=handle_reply)

    def delete_cookie(self, name: str) -> None:
        """Delete given cookie if cookie exists."""
        self.web_session.clear_cookies(self.endpoints['webHost'], '/', name)

    def set_cookie(self, name: str, value: str) -> None:
        """Set cookie value in Cookie jar. Defaults cookie domain as the WEB_HOST."""
        self.web_session.set_cookie(name, value, self.endpoints['webHost'])

    @property
    def token(self) -> str | None:
        """343 auth token to authenticate self to API endpoints.

        :raises ValueError: If the value being set doesn't contain the version identifier ("v4=").
        """
        return self._token

    @token.setter
    def token(self, value: str) -> None:
        value = decode_url(value)
        if (key_start_index := value.find('v4=')) == -1:
            raise ValueError('token value is missing version identifier ("v4=") to signify start.')

        self._token = value[key_start_index:].rstrip()
        self.set_cookie('343-spartan-token', self._token)
        self.api_session.headers['x-343-authorization-spartan'] = self._token

        hide_windows_file(HI_TOKEN_PATH, unhide=True)
        HI_TOKEN_PATH.write_text(self._token, encoding='utf8')
        hide_windows_file(HI_TOKEN_PATH)

    @token.deleter
    def token(self) -> None:
        self._token = None
        self.delete_cookie('343-spartan-token')
        if 'x-343-authorization-spartan' in self.api_session.headers:
            self.api_session.headers.pop('x-343-authorization-spartan')
        if HI_TOKEN_PATH.is_file():
            HI_TOKEN_PATH.unlink()

    @property
    def wpauth(self) -> str | None:
        """Halo Waypoint auth key to create 343 spartan tokens."""
        return self._wpauth

    @wpauth.setter
    def wpauth(self, value: str) -> None:
        value = decode_url(value)
        self._wpauth = value.split(':')[-1].strip()
        self.set_cookie('wpauth', self._wpauth)

        hide_windows_file(HI_WPAUTH_PATH, unhide=True)
        HI_WPAUTH_PATH.write_text(self._wpauth, encoding='utf8')
        hide_windows_file(HI_WPAUTH_PATH)

    @wpauth.deleter
    def wpauth(self) -> None:
        self._wpauth = None
        self.delete_cookie('wpauth')
        if HI_WPAUTH_PATH.is_file():
            HI_WPAUTH_PATH.unlink()

    @property
    def api_root(self) -> str:
        """Root of sent API requests."""
        return f'https://{self.host}{self.parent_path}'

    @property
    def host(self) -> str:
        """Host to send to."""
        return f'{self.sub_host}.{self.endpoints["svcHost"]}'

    @host.setter
    def host(self, value: str) -> None:
        self.sub_host = value



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""File menu implementation."""
from __future__ import annotations

__all__ = (
    'export_data',
    'FileContextMenu',
    'import_data',
)

from pathlib import Path

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...constants import *
from ...models import DeferredCallable
from ...utils import add_menu_items
from ...utils import format_tb
from ...utils import init_objects
from ..aliases import app
from ..aliases import tr
from ..workers import ExportData
from ..workers import ImportData

# Default file extension is .7z
_ARCHIVE_FILTER: str = ';;'.join((
    'Archive Files (*.7z *.zip *.piz *.tar *.tar.gz *.tgz *.tar.bz2 *.tbz2 *.tar.xz *.txz)',
    'All files (*.*)'
))


def export_data() -> None:
    """Export cached_requests to an archive file selected by a :py:class:`QFileDialog`.

    The following archive types are supported::

        7z, tar, zip, gztar, bztar, xztar
    """
    export_dest = Path(QFileDialog.getSaveFileName(None, caption=tr('gui.menus.file.export'),  # type: ignore
                                                   dir=str(Path.home()), filter=_ARCHIVE_FILTER)[0])
    # Return early if no file selected
    if export_dest.is_dir():
        return

    if ExportData.EXTENSION_FORMATS.get(export_dest.suffix) is None:
        app().show_dialog('errors.unsupported_archive_type', description_args=(export_dest, export_dest.suffix,))
        return

    app().start_worker(ExportData(
        base=HI_CACHE_PATH / 'cached_requests',
        dest=export_dest,
        exceptionRaised=lambda e: app().show_dialog(
            'errors.cache_export_failure', description_args=(export_dest, e,),
            details_text=format_tb(e.__traceback__)
        )))


def import_data() -> None:
    """Import data from an archive file selected by a :py:class:`QFileDialog`.

    The following archive extensions are supported::

        .7z .zip .piz .tar .tar.gz .tgz .tar.bz2 .tbz2 .tar.xz .txz
    """
    archive_file = Path(QFileDialog.getOpenFileName(None, caption=tr('gui.menus.file.import'),  # type: ignore
                                                    dir=str(Path.home() / 'Downloads'), filter=_ARCHIVE_FILTER)[0])
    # Return early if no file selected
    if archive_file.is_dir():
        return

    # Attempt to unpack file into temp dir
    app().start_worker(ImportData(
        archive=archive_file,
        dest=HI_CACHE_PATH,
        exceptionRaised=lambda e: app().show_dialog(
            'errors.cache_import_failure', description_args=(archive_file, e,),
            details_text=format_tb(e.__traceback__)
        )))


class FileContextMenu(QMenu):
    """Context menu that shows actions to manipulate files."""

    def __init__(self, parent) -> None:
        """Create a new :py:class:`FileContextMenu`."""
        super().__init__(parent)

        cached_requests = HI_CACHE_PATH / 'cached_requests'

        init_objects({
            (open_explorer := QAction(self)): {
                'text': tr('gui.menus.file.open'),
                'icon': (app().get_theme_icon('dialog_open') or
                         app().icon_store['folder']),
                'triggered': DeferredCallable(
                    QDesktopServices.openUrl,
                    QUrl(HI_CACHE_PATH.as_uri())
                )
            },

            (flush_cache := QAction(self)): {
                # DISABLED IF EMPTY DIRECTORY
                'disabled': not any(HI_CACHE_PATH.iterdir()),
                'text': tr('gui.menus.file.flush'),
                'icon': (app().get_theme_icon('dialog_discard') or
                         self.style().standardIcon(QStyle.StandardPixmap.SP_DialogDiscardButton)),
                'triggered': self.flush_cache
            },

            (import_from := QAction(self)): {
                'text': tr('gui.menus.file.import'),
                'icon': app().icon_store['import'],
                'triggered': import_data
            },

            (export_to := QAction(self)): {
                # DISABLED IF EMPTY DIRECTORY OR NOT DIRECTORY
                'disabled': (True if (cached_requests.is_dir() and not any(cached_requests.iterdir()))
                             else not cached_requests.is_dir()),
                'text': tr('gui.menus.file.export'),
                'icon': app().icon_store['export'],
                'triggered': export_data
            }
        })

        add_menu_items(self, [
            'Files', open_explorer, flush_cache,
            'Import/Export', import_from, export_to
        ])

    def flush_cache(self) -> None:
        """Remove all data from cache."""
        do_flush: bool = app().show_dialog(
            'warnings.delete_cache', self,
            QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel,
            default_button=QMessageBox.StandardButton.Cancel
        ).role == QMessageBox.ButtonRole.AcceptRole

        if do_flush:
            app().client.searched_paths.clear()
            QDir(HI_CACHE_PATH).removeRecursively()
            HI_CACHE_PATH.mkdir(exist_ok=True)



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Custom :py:class:`Exception`'s and excepthook implementation."""
from __future__ import annotations

__all__ = (
    'ExceptHookCallable',
    'ExceptionEvent',
    'ExceptionHook',
)

import sys
from collections.abc import Callable
from types import TracebackType
from typing import TypeAlias

from .events import Event
from .events import EventBus

ExceptHookCallable: TypeAlias = Callable[[type[BaseException], BaseException, TracebackType], None]


class ExceptionEvent(Event):
    """Event fired when an exception is caught by an :py:class:`ExceptionHook`."""

    __slots__ = ('exception', 'traceback')

    def __init__(self, exception: BaseException, traceback: TracebackType) -> None:
        """Create a new :py:class:`ExceptionEvent` event with the given exception and its associated traceback."""
        self.exception: BaseException = exception
        self.traceback: TracebackType = traceback


class ExceptionHook:
    """Object that intercepts :py:class:`Exception`'s and handles them."""

    def __init__(self, bus_id: str | None = 'exceptions'):
        """Initialize the :py:class:`ExceptionHook` for use in a context manager."""
        self.__old_hook: ExceptHookCallable = sys.excepthook
        self.event_bus: EventBus = EventBus(bus_id)

    def __call__(self, type_: type[BaseException], exception: BaseException, traceback: TracebackType) -> None:
        """When an exception is raised."""
        # Don't handle BaseExceptions
        if not issubclass(type_, Exception):
            return self.old_hook(type_, exception, traceback)

        self.event_bus << ExceptionEvent(exception, traceback)

    def __repr__(self) -> str:
        """Representation of the :py:class:`ExceptionHook` with the old hook."""
        return f'<{type(self).__name__} ({self.old_hook=})>'

    def __enter__(self) -> None:
        """Temporary extend current exception hook."""
        sys.excepthook = self

    def __exit__(self, *_) -> None:
        """Reset current exception hook to the original one."""
        sys.excepthook = self.old_hook
        del EventBus['exceptions']

    @property
    def old_hook(self) -> ExceptHookCallable:
        """Return the original exception hook."""
        return self.__old_hook



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Package containing miscellaneous :py:class:`QWidget` Widgets."""

__all__ = (
    'CacheExplorer',
    'ComboBox',
    'ExceptionLogger',
    'ExternalTextBrowser',
    'HistoryComboBox',
    'PasteLineEdit',
    'TranslatableComboBox',
)

from .cache_explorer import CacheExplorer
from .combo_box import ComboBox
from .combo_box import HistoryComboBox
from .combo_box import TranslatableComboBox
from .exception_logger import ExceptionLogger
from .external_text_browser import ExternalTextBrowser
from .paste_line_edit import PasteLineEdit



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Utilities for networking."""
from __future__ import annotations

__all__ = (
    'decode_url',
    'dict_to_cookie_list',
    'dict_to_query',
    'encode_url_params',
    'guess_json_utf',
    'http_code_map',
    'is_error_status',
    'query_to_dict',
    'wait_for_reply',
)

import codecs
from http import HTTPStatus
from urllib.parse import unquote as decode_url
from urllib.parse import urlencode as encode_url_params

from PySide6.QtCore import *
from PySide6.QtNetwork import *

# pylint: disable=not-an-iterable
http_code_map = {status.value: (status.phrase, status.description) for status in HTTPStatus}
for _value, _description in {
        400: 'Your search path has malformed syntax or bad characters.',
        401: 'No permission -- Your API token is most likely invalid.',
        403: 'Request forbidden -- You cannot get this resource with or without an API token.',
        404: 'No resource found at the given location.',
        405: 'Invalid method -- GET requests are not accepted for this resource.',
        406: 'Client does not support the given resource format.',
}.items():
    http_code_map[_value] = (http_code_map[_value][0], _description)
    del _value, _description


def dict_to_cookie_list(cookie_values: dict[str, str]) -> list[QNetworkCookie]:
    """Transform a name and value pair into a list of :py:class:`QNetworkCookie` objects."""
    return [QNetworkCookie(
        name=name.encode('utf8'),
        value=value.encode('utf8')
    ) for name, value in cookie_values.items()]


# noinspection PyTypeChecker
def dict_to_query(params: dict[str, str]) -> QUrlQuery:
    """Transform a param name and value pair into a :py:class:`QUrlQuery` object."""
    query = QUrlQuery()
    query.setQueryItems(list(params.items()))
    return query


# noinspection PyTypeChecker
def query_to_dict(query: QUrlQuery | str) -> dict[str, str]:
    """Translate a query string with the format of QUrl.query() to a dictionary representation."""
    if isinstance(query, QUrlQuery):
        query = query.query()
    query = query.lstrip('?')

    return {} if not query else dict(
        pair.split('=') for pair in query.split('&')
    )


def is_error_status(status: int) -> bool:
    """Return True if the HTTP status code is an error status."""
    return 400 <= status < 600


def wait_for_reply(reply: QNetworkReply) -> None:
    """Process events until the reply is finished.

    :param reply: The QNetworkReply to wait for.
    :raises RuntimeError: If the internal C++ QNetworkRequest is deleted.
    """
    while not reply.isFinished():
        QCoreApplication.processEvents()

##########
# NOTICE:
##########
# Requests
# Copyright 2019 Kenneth Reitz
# Apache 2.0 License
# https://github.com/psf/requests/blob/main/LICENSE


# pylint: disable=consider-using-assignment-expr
def guess_json_utf(data: bytes) -> str | None:
    """:return: String representing the detected encoding of the given data. None if not detected."""
    # JSON always starts with two ASCII characters, so detection is as
    # easy as counting the nulls and from their location and count
    # determine the encoding. Also detect a BOM, if present.
    sample: bytes = data[:4]

    if sample in (codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE):
        return 'utf-32'     # BOM included
    if sample[:3] == codecs.BOM_UTF8:
        return 'utf-8-sig'  # BOM included, MS style (discouraged)
    if sample[:2] in (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE):
        return 'utf-16'     # BOM included

    match sample.count(b'\x00'):
        case 0:
            return 'utf-8'

        case 2:
            if sample[::2] == b'\x00\x00':   # 1st and 3rd are null
                return 'utf-16-be'
            if sample[1::2] == b'\x00\x00':  # 2nd and 4th are null
                return 'utf-16-le'

        case 3:
            if sample[:3] == b'\x00\x00\x00':  # First 3 are null
                return 'utf-32-be'
            if sample[1:] == b'\x00\x00\x00':  # Last 3 are null
                return 'utf-32-le'

    return None



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""A simple way to get live Halo data straight from Halo Waypoint."""
# TODO: Add logging functionality
# TODO: Redo Events implementation
# TODO: Remove pyright comments when PySide6 signals and events are properly type hinted.
from ._version import __version__
from ._version import __version_info__

__author__ = 'Cubicpath@Github <cubicpath@pm.me>'
"""Author's information."""



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""System utility functions."""
from __future__ import annotations

__all__ = (
    'create_shortcut',
    'get_desktop_path',
    'get_start_menu_path',
    'get_winreg_value',
    'hide_windows_file',
    'patch_windows_taskbar_icon',
)

from collections.abc import Callable
from pathlib import Path
from typing import Any

from ..constants import *
from .common import bit_rep
from .common import quote_str


def create_shortcut(target: Path, arguments: str | None = None,
                    name: str | None = None, description: str | None = None,
                    icon: Path | None = None, working_dir: Path | None = None,
                    desktop: bool = True, start_menu: bool = True,
                    version: str | None = None, terminal: bool = True) -> None:
    """Create a shortcut on the given path.

    Notes
        * start_menu is ignored on macOS
        * terminal is ignored by Windows
        * working_dir is Windows only
        * version is Linux only

    Linux and macOS implementations are heavily based on pyshortcuts.

    :param target: Target of the shortcut.
    :param arguments: Command line arguments to pass to the target.
    :param version: Version identifier of the target.
    :param terminal: Whether to open the target with a terminal.
    :param name: Name of the shortcut.
    :param description: Description of the shortcut.
    :param icon: Path to an icon to use for the shortcut.
    :param working_dir: Working directory to start in when executing the shortcut.
    :param desktop: Whether to create a desktop shortcut.
    :param start_menu: Whether to create a start menu shortcut.
    :raises ValueError: icon extension cannot be used as an icon for the given platform.
    """
    import shutil
    import subprocess
    import sys

    if not desktop and not start_menu:
        return

    target = target.resolve(True).absolute()
    name = 'Shortcut' if name is None else name
    working_dir = Path.home() if working_dir is None else working_dir

    PLATFORM_SHORTCUT_DATA: dict[str, dict[str, Any]] = {
        'darwin': {
            'shortcut_ext': '.app',
            'icon_exts': ('.icns',),
        },
        'linux': {
            'shortcut_ext': '.desktop',
            'icon_exts': ('.ico', '.svg', '.png'),
        },
        'win32': {
            'shortcut_ext': '.lnk',
            'icon_exts': ('.ico', '.exe'),
        }
    }

    if not (data := PLATFORM_SHORTCUT_DATA.get(sys.platform)):
        return

    if icon and icon.suffix not in data['icon_exts']:
        raise ValueError(f'Icon must be one of {data["icon_exts"]} for {sys.platform}')

    if (platform := sys.platform.lower()) == 'darwin':
        # macOS doesn't support start menu shortcuts, so return if not creating a desktop shortcut
        if not desktop:
            return

        # Create the desktop directory if it doesn't exist
        if (desktop_path := get_desktop_path()) is not None:
            if not desktop_path.is_dir():
                desktop_path.mkdir(parents=True)

            # Create the shortcut folders, replacing if it already exists
            dest = (desktop_path / name).with_suffix(data['shortcut_ext'])
            if dest.exists():
                shutil.rmtree(dest)

            dest.mkdir(parents=True)
            (dest / 'Contents').mkdir()
            (dest / 'Contents/MacOS').mkdir()
            (dest / 'Contents/Resources').mkdir()

            # Add macOS shortcut data
            with (dest / 'Contents/Info.plist').open('w', encoding='utf8') as plist:
                # noinspection HttpUrlsUsage
                plist.writelines([
                    '<?xml version="1.0" encoding="UTF-8"?>\n',
                    '<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN"\n',
                    '"http://www.apple.com/DTDs/PropertyList-1.0.dtd">\n',
                    '<plist version="1.0">\n',
                    '  <dict>\n',
                    f'  <key>CFBundleGetInfoString</key> <string>{description or ""}</string>\n',
                    f'  <key>CFBundleName</key> <string>{name}</string>\n',
                    f'  <key>CFBundleExecutable</key> <string>{name}</string>\n',
                    f'  <key>CFBundleIconFile</key> <string>{name}</string>\n',
                    '  <key>CFBundlePackageType</key> <string>APPL</string>\n',
                    '  </dict>\n',
                    '</plist>\n',
                ])

            with (dest / f'Contents/MacOS/{name}').open('w', encoding='utf8') as shortcut_script:
                shortcut_script.writelines([
                    '#!/bin/bash\n',
                    # These exports are not used if the script is ran from the terminal
                    f'export SCRIPT={target}\n',
                ])
                if arguments is not None:
                    shortcut_script.write(f'export ARGS=\'{arguments}\'\n')

                if not terminal:
                    shortcut_script.write(f'$SCRIPT{" $ARGS" if arguments else ""}')
                else:
                    osa_script = f'{target} {arguments}'.replace(' ', '\\ ')
                    shortcut_script.writelines([
                        'osascript - e \'tell application "Terminal"\n',
                        f'do script "\'{osa_script}\'"\n',
                        'end tell\n',
                        '\'\n',
                    ])

                shortcut_script.write('\n')

            # Change permissions to allow execution
            (dest / f'Contents/MacOS/{name}').chmod(0o755)  # rwxr-xr-x

            # Add the icon
            if icon:
                shutil.copy(icon, (dest / f'Contents/Resources/{name}').with_suffix(icon.suffix))

    elif platform.startswith('linux'):
        entry_values: dict[str, object] = {
            'Encoding': 'UTF-8',
            'Version': version,
            'Type': 'Application',
            'Exec': f'{target} {arguments}',
            'Terminal': terminal,
            'Icon': icon,
            'Name': name,
            'Comment': description,
        }

        for (do, path) in (
                (desktop, get_desktop_path()),
                (start_menu, get_start_menu_path())
        ):
            if do:
                # Create the directory if it doesn't exist
                if not path.is_dir():
                    path.mkdir(parents=True)

                # Create the .desktop file
                dest = (path / name).with_suffix(data['shortcut_ext'])
                with dest.open('w', encoding='utf8') as shortcut_script:
                    shortcut_script.write('[Desktop Entry]\n')
                    shortcut_script.writelines([
                        f'{k}={v}\n' for k, v in entry_values.items() if v is not None
                    ])

                # Change permissions to allow execution
                dest.chmod(0o755)  # rwxr-xr-x

    elif platform == 'win32':
        arg_factories: dict[str, tuple[Any, Callable[[Any], str]]] = {
            'Target': (target, quote_str),
            'Arguments': (arguments, quote_str),
            'Name': (name, quote_str),
            'Description': (description, quote_str),
            'Icon': (icon, quote_str),
            'WorkingDirectory': (working_dir, quote_str),
            'Extension': (data['shortcut_ext'], quote_str),
            'Desktop': (desktop, bit_rep),
            'StartMenu': (start_menu, bit_rep)
        }

        abs_script_path: Path = (HI_RESOURCE_PATH / 'scripts/CreateShortcut.ps1').resolve(True).absolute()
        powershell_arguments = [
            'powershell.exe', '-ExecutionPolicy', 'Unrestricted', abs_script_path,
        ]

        # Append keyword arguments to the powershell script if the value is not None
        # Every argument is in the form of -<keyword>:<value> with value being
        # represented as a quoted string or raw integer.
        powershell_arguments.extend([
            f'-{key}:{factory(value)}' for (key, (value, factory)) in arg_factories.items() if value is not None
        ])

        # All user input is passed in as a controlled powershell.exe argument.
        subprocess.run(  # nosec B603:subprocess_without_shell_equals_true
            powershell_arguments, capture_output=True, text=True, check=True
        )


def get_desktop_path() -> Path:
    """Cross-platform utility to obtain the path to the desktop.

    This function is cached after the first call.

    * On Windows, this returns the path found in the registry, or
      the default ~/Desktop if the registry could not be read from.

    * On Linux and macOS, this returns the DESKTOP value in ~/.config/user-dirs.dirs file, or
      the default ~/Desktop.

    :return: Path to the user's desktop directory.
    :raises TypeError: If winreg value for Desktop shell folder is not a string/path.
    """
    import os
    import sys

    # Assume that once found, the desktop path does not change
    if hasattr(get_desktop_path, '__cached__'):
        return get_desktop_path.__cached__

    platform: str = sys.platform.lower()
    desktop: Path

    if platform == 'win32':
        shell_folder_key = r'HKEY_CURRENT_USER\SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\User Shell Folders'
        desktop = Path.home() / 'Desktop'

        try:
            val = get_winreg_value(shell_folder_key, 'Desktop')
            if not isinstance(val, str):
                raise TypeError(f'Winreg value of {shell_folder_key}\\Desktop is not a string.')

            # Make sure the path is resolved
            desktop = Path(val).resolve(True).absolute()

        except (ImportError, OSError):
            pass  # Return the default windows path if the registry couldn't be read.

    else:  # Linux, Darwin
        home: Path = Path.home() or Path(os.environ['HOME'])
        desktop = home / 'Desktop'

        # If desktop is defined in user's config, use that
        dir_file: Path = home / '.config/user-dirs.dirs'
        if dir_file.is_file():
            with dir_file.open(mode='r', encoding='utf8') as f:
                text: list[str] = f.readlines()

            for line in text:
                # Read the DESKTOP variable's value and evaluate it
                if 'DESKTOP' in line:
                    line: str = line.replace('$HOME', str(home))[:-1]
                    config_val: str = line.split('=')[1].strip('\'\"')
                    desktop = Path(config_val).resolve(True).absolute()

    get_desktop_path.__cached__ = desktop
    return desktop


def get_start_menu_path() -> Path:
    """Cross-platform utility to obtain the path to the Start Menu or equivalent.

    This function is cached after the first call.

    * On Windows, this returns the main Start Menu folder, so it is
      recommended that you use the "Programs" sub-folder for adding shortcuts.

    * On Linux, this returns the ~/.local/share/applications directory.

    * On macOS, this raises a FileNotFoundError.

    :return: Path to the Start Menu or None if not found.
    :raises FileNotFoundError: If attempted on macOS.
    :raises TypeError: If winreg value for Start Menu shell folder is not a string/path.
    """
    import os
    import sys

    # Assume that once found, the start menu path does not change
    if hasattr(get_start_menu_path, '__cached__'):
        return get_start_menu_path.__cached__

    platform: str = sys.platform.lower()
    start_menu: Path

    if platform == 'darwin':
        raise FileNotFoundError('macOS has no Start Menu folder.')

    if platform == 'win32':
        shell_folder_key = r'HKEY_CURRENT_USER\SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\User Shell Folders'
        start_menu = Path.home() / 'AppData/Roaming/Microsoft/Windows/Start Menu'

        try:
            val = get_winreg_value(shell_folder_key, 'Start Menu')
            if not isinstance(val, str):
                raise TypeError(f'Winreg value of {shell_folder_key}\\Start Menu is not a string.')

            # Make sure the path is resolved
            start_menu = Path(val).resolve(True).absolute()

        except (ImportError, OSError):
            pass  # Return the default windows path if the registry couldn't be read.

    else:  # Linux
        home: Path = Path.home() or Path(os.environ['HOME'])

        start_menu = home / '.local/share/applications'

    get_start_menu_path.__cached__ = start_menu
    return start_menu


def get_winreg_value(key_name: str, value_name: str) -> str | int | bytes | list | None:
    """Get a value from the Windows registry.

    :param key_name: The registry key to read. The parent key must be the name of a defined winreg constant.
    :param value_name: The value to read.
    :return: The value, or None if not found.
    :raises AttributeError: If the parent_key is not a defined winreg constant.
    :raises ImportError: If winreg is not available.
    :raises OSError: If the registry key could not be read.
    """
    # noinspection PyCompatibility
    import winreg
    from os.path import expandvars

    parent_key: int = getattr(winreg, key_name.split('\\')[0])
    sub_key: str = '\\'.join(key_name.split('\\')[1:])
    if not isinstance(parent_key, int):
        raise AttributeError('parent_key is not a defined winreg constant.')

    reg_key = winreg.OpenKey(parent_key, sub_key, 0, winreg.KEY_QUERY_VALUE)
    val, reg_type = winreg.QueryValueEx(reg_key, value_name)

    reg_key.Close()

    # Expand environment variables
    if reg_type == winreg.REG_EXPAND_SZ:
        val = expandvars(val)

    return val


# pylint: disable=R
def hide_windows_file(file_path: Path | str, *, unhide: bool = False) -> int | None:
    """Hide an existing Windows file. If not running windows, do nothing.

    Use unhide kwarg to reverse the operation

    :param file_path: Absolute or relative path to hide.
    :param unhide: Unhide a hidden file in Windows.
    :return: None if not on Windows, else if the function succeeds, the return value is nonzero.
    """
    import sys

    # Resolve string path to use with kernel32
    file_path = str(Path(file_path).resolve())
    if sys.platform == 'win32':
        from ctypes import windll

        # File flags are a 32-bit bitarray, the "hidden" attribute is the 2nd least significant bit
        FILE_ATTRIBUTE_HIDDEN = 0b00000000000000000000000000000010

        # bitarray for boolean flags representing file attributes
        current_attributes: int = windll.kernel32.GetFileAttributesW(file_path)
        if not unhide:
            # Add hide attribute to bitarray using bitwise OR
            # 0b00000000 -> 0b00000010 ---- 0b00000110 -> 0b00000110
            merged_attributes: int = current_attributes | FILE_ATTRIBUTE_HIDDEN
            return windll.kernel32.SetFileAttributesW(file_path, merged_attributes)
        else:
            # Remove hide attribute from bitarray if it exists
            # Check with bitwise AND; Remove with bitwise XOR
            # 0b00000100 -> 0b00000100 ---- 0b00000110 -> 0b00000100
            # Only Truthy returns (which contain the hidden attribute) will subtract from the bitarray
            is_hidden = bool(current_attributes & FILE_ATTRIBUTE_HIDDEN)
            if is_hidden:
                subtracted_attributes: int = current_attributes ^ FILE_ATTRIBUTE_HIDDEN
                return windll.kernel32.SetFileAttributesW(file_path, subtracted_attributes)


def patch_windows_taskbar_icon(app_id: str = '') -> int | None:
    """Override Python's default Windows taskbar icon with the custom one set by the app window.

    See:
    https://docs.microsoft.com/en-us/windows/win32/api/shobjidl_core/nf-shobjidl_core-setcurrentprocessexplicitappusermodelid
    for more information.

    :param app_id: Pointer to the AppUserModelID to assign to the current process.
    :return: None if not on Windows, S_OK if this function succeeds. Otherwise, it returns an HRESULT error code.
    """
    import sys

    if sys.platform == 'win32':
        from ctypes import windll
        return windll.shell32.SetCurrentProcessExplicitAppUserModelID(app_id)



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Implementations for worker-thread runnables."""
from __future__ import annotations

__all__ = (
    'ExportData',
    'ImportData',
)

import shutil
from collections.abc import Callable
from pathlib import Path

from PySide6.QtCore import *
from shiboken6 import Shiboken


class _SignalHolder(QObject):
    exceptionRaised = Signal(Exception)
    valueReturned = Signal(object)


class _Worker(QRunnable):
    _signal_holder: type[_SignalHolder] = _SignalHolder

    def __init__(self, **kwargs: Callable | Slot) -> None:
        super().__init__()
        self.signals = self._signal_holder()

        # Connect signals from keyword arguments
        for kw, val in kwargs.items():
            if hasattr(self.signals, kw) and isinstance((signal := getattr(self.signals, kw)), SignalInstance):
                signal.connect(val)
            else:
                raise TypeError(f'"{kw}" is not a valid kwarg or signal name.')

    # pylint: disable=no-self-use
    def _dummy_method(self) -> None:
        return

    def _run(self) -> None:
        raise NotImplementedError

    # pylint: disable=broad-except
    @Slot()
    def run(self) -> None:
        """Ran by the :py:class:`QThreadPool`.

        Sends non-``None`` return values through the ``valueReturned`` signal.
            - If you need to return ``None``, I suggest creating a separate object to represent it.

        Sends any uncaught :py:class:`Exception`'s through the ``exceptionRaised`` signal.

        :raises RuntimeError: If worker started before application instance is defined.
        """
        if (app := QCoreApplication.instance()) is None:
            raise RuntimeError('Worker started before application instance is defined.')

        # Workaround for application deadlock (issue #31)
        app.aboutToQuit.connect(  # pyright: ignore[reportGeneralTypeIssues]
            self._dummy_method, Qt.ConnectionType.BlockingQueuedConnection
        )

        try:
            # Emit non-None return values through the `valueReturned` signal.
            if (ret_val := self._run()) is not None:
                self.signals.valueReturned.emit(ret_val)

        except Exception as e:
            # This occurs when the application is exiting and an internal C++ object is being read after it is deleted.
            # So, return quietly and allow the process to exit with no errors.
            if not Shiboken.isValid(self.signals):
                return

            self.signals.exceptionRaised.emit(e)

        # Disconnect deadlock safeguard if successful to avoid possible leak
        app.aboutToQuit.disconnect(self._dummy_method)  # pyright: ignore[reportGeneralTypeIssues]

        # Delete if not deleted
        if Shiboken.isValid(self.signals):
            self.signals.deleteLater()


class ExportData(_Worker):
    """Build archive in a temp file then remove .tmp extension when finished."""

    EXTENSION_FORMATS: dict[str, str] = {
        '.7z': '7z',
        '.tar': 'tar',
        '.zip': 'zip', '.piz': 'zip',
        '.tar.gz': 'gztar', '.tgz': 'gztar',
        '.tar.bz2': 'bztar', '.tbz2': 'bztar',
        '.tar.xz': 'xztar', '.txz': 'xztar'
    }

    def __init__(self, base: Path, dest: Path, **kwargs: Callable | Slot) -> None:
        """Create a new :py:class:`ExportData` worker to run with the given ``base`` path and ``dest`` file."""
        super().__init__(**kwargs)
        self.base = base
        self.dest = dest

    def _run(self) -> None:
        # This implicitly raises ValueError if the destination suffix is not a valid extension format.
        temp_archive_path: str = shutil.make_archive(
            base_name=QDir.tempPath() + f'/{self.dest.stem}',
            format=self.EXTENSION_FORMATS.get(self.dest.suffix, f'.{self.dest.suffix}'),
            root_dir=self.base.parent,
            base_dir=f'./{self.base.name}'
        )

        shutil.move(
            src=temp_archive_path,
            dst=self.dest
        )


class ImportData(_Worker):
    """Extract from archive into a directory."""

    def __init__(self, archive: Path, dest: Path, **kwargs: Callable | Slot) -> None:
        """Create a new :py:class:`ImportData` worker to run with the given ``archive`` file and ``dest`` path."""
        super().__init__(**kwargs)
        self.archive = archive
        self.dest = dest

    def _run(self) -> None:
        shutil.unpack_archive(str(self.archive), extract_dir=self.dest)



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Utility functions for hi_getter."""
from __future__ import annotations

__all__ = (
    'add_menu_items',
    'bit_rep',
    'create_shortcut',
    'current_requirement_licenses',
    'current_requirement_names',
    'current_requirement_versions',
    'decode_url',
    'delete_layout_widgets',
    'dict_to_cookie_list',
    'dict_to_query',
    'dump_data',
    'encode_url_params',
    'format_tb',
    'get_desktop_path',
    'get_parent_doc',
    'get_start_menu_path',
    'get_winreg_value',
    'guess_json_utf',
    'has_package',
    'hide_windows_file',
    'http_code_map',
    'icon_from_bytes',
    'init_layouts',
    'init_objects',
    'is_error_status',
    'patch_windows_taskbar_icon',
    'query_to_dict',
    'quote_str',
    'return_arg',
    'scroll_to_top',
    'set_or_swap_icon',
    'unique_values',
    'wait_for_reply',
)

from .common import bit_rep
from .common import dump_data
from .common import format_tb
from .common import get_parent_doc
from .common import return_arg
from .common import quote_str
from .common import unique_values
from .gui import add_menu_items
from .gui import delete_layout_widgets
from .gui import icon_from_bytes
from .gui import init_layouts
from .gui import init_objects
from .gui import scroll_to_top
from .gui import set_or_swap_icon
from .network import decode_url
from .network import dict_to_cookie_list
from .network import dict_to_query
from .network import encode_url_params
from .network import guess_json_utf
from .network import http_code_map
from .network import is_error_status
from .network import query_to_dict
from .network import wait_for_reply
from .package import current_requirement_licenses
from .package import current_requirement_names
from .package import current_requirement_versions
from .package import has_package
from .system import create_shortcut
from .system import get_desktop_path
from .system import get_start_menu_path
from .system import get_winreg_value
from .system import hide_windows_file
from .system import patch_windows_taskbar_icon



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Tools menu implementation."""
from __future__ import annotations

__all__ = (
    'create_app_shortcut',
    'ToolsContextMenu',
)

import sys
from pathlib import Path

from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...constants import *
from ...utils import add_menu_items
from ...utils import init_objects
from ...utils import create_shortcut
from ..aliases import app
from ..aliases import tr


def create_app_shortcut() -> None:
    """Create shortcut for starting program."""
    exec_path = Path(sys.executable)

    desktop_button = QPushButton(tr('gui.menus.tools.create_shortcut.only_desktop'))
    start_menu_button = QPushButton(tr('gui.menus.tools.create_shortcut.only_start_menu'))
    both_button = QPushButton(tr('gui.menus.tools.create_shortcut.both'))

    # Show dialog, return early if user presses cancel
    if (response := app().show_dialog(
        'questions.create_shortcut', None, (
            (desktop_button, QMessageBox.ButtonRole.AcceptRole),
            (start_menu_button, QMessageBox.ButtonRole.AcceptRole),
            (both_button, QMessageBox.ButtonRole.AcceptRole),
            QMessageBox.StandardButton.Cancel
        ), default_button=QMessageBox.StandardButton.Cancel
    )).role == QMessageBox.ButtonRole.RejectRole:
        return

    # If response is affirmative, mark which shortcuts to create
    do_desktop = response.button in {desktop_button, both_button}
    do_start_menu = response.button in {start_menu_button, both_button}

    # Get the windowless Python executable name (append w)
    name = exec_path.stem
    shortcut_path = exec_path.with_stem(name if name.endswith('w') else name + 'w')

    # Create shortcut to launch this package, with proper kwargs
    create_shortcut(target=shortcut_path, arguments=f'-m {HI_PACKAGE_NAME}',
                    name=tr('app.name'), description=tr('app.description'),
                    icon=HI_RESOURCE_PATH / 'icons/hi.ico', terminal=False,
                    desktop=do_desktop, start_menu=do_start_menu)


class ToolsContextMenu(QMenu):
    """Context menu for tools."""

    def __init__(self, parent) -> None:
        """Create a new :py:class:`ToolsContextMenu`."""
        super().__init__(parent)

        # noinspection PyUnresolvedReferences
        init_objects({
            (scan_selector := QAction(self)): {
                'text': tr('gui.menus.tools.scan_selector_dialog'),
                'icon': self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogContentsView),
                'triggered': app().windows['scan_selector'].show
            },

            (shortcut_tool := QAction(self)): {
                'text': tr('gui.menus.tools.create_shortcut'),
                'icon': self.style().standardIcon(QStyle.StandardPixmap.SP_DesktopIcon),
                'triggered': create_app_shortcut
            },

            (exception_reporter := QAction(self)): {
                'text': tr('gui.menus.tools.exception_reporter'),
                'icon': app().windows['app'].exception_reporter.logger.icon(),  # type: ignore
                'triggered': app().windows['app'].exception_reporter.show       # type: ignore
            }
        })

        add_menu_items(self, [
            'Network', scan_selector,
            'Other', shortcut_tool, exception_reporter
        ])



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Utilities for GUI elements."""
from __future__ import annotations

__all__ = (
    'add_menu_items',
    'delete_layout_widgets',
    'icon_from_bytes',
    'init_layouts',
    'init_objects',
    'scroll_to_top',
    'set_or_swap_icon',
)

from collections.abc import Iterable
from collections.abc import Sequence
from typing import Any
from typing import Literal
from typing import TypeAlias

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

_LAYOUT_OBJ: TypeAlias = QLayout | QLayoutItem | QWidget

_layout_type_map = {
    QLayout: 'addLayout',
    QLayoutItem: 'addItem',
    QWidget: 'addWidget',
}

_menu_type_map = {
    str: QMenu.addSection,
    QMenu: QMenu.addMenu,
    QAction: QMenu.addAction,
}


def add_menu_items(menu: QMenu, items: Sequence[str | QAction | QMenu]) -> None:
    """Add items to the given menu.

    This uses the associated :py:class:`QMenu` methods for each object's type::

        str: QMenu.addSection,
        QMenu: QMenu.addMenu,
        QAction: QMenu.addAction,


    :param menu: The menu to add items to.
    :param items: The items to add to the menu.
    """
    for obj in items:
        # Find the item's type and associated method.
        for item_type, meth in _menu_type_map.items():

            if isinstance(obj, item_type):
                # Run method and go to next item.
                meth(menu, obj)
                break


def delete_layout_widgets(layout: QLayout) -> None:
    """Delete all widgets in a layout."""
    while (item := layout.takeAt(0)) is not None:
        item.widget().deleteLater()


def icon_from_bytes(data: bytes) -> QIcon:
    """Create a :py:class:`QIcon` from bytes using a :py:class:`QPixmap` as a proxy."""
    pixmap = QPixmap()
    pixmap.loadFromData(data)
    icon = QIcon(pixmap)
    return icon


def init_layouts(layout_data: dict[
    QLayout,
    dict[
        Literal['childLayouts', 'childWidgets', 'items'],
        Sequence[_LAYOUT_OBJ | Sequence]
    ]
]) -> None:
    """Initialize :py:class:`QLayout` hierarchies with the given data.

    The only secondary dictionary keys allowed are ``'childLayouts'``, ``'childWidgets'``, and ``'items'``.

    layout_data should be a dictionary structured like this::

        {
            my_layout1: {},
            my_layout2: {'items': [layout1, widget1, my_layout1, widget2, layout2, widget3]},
            my_layout3: {'childLayouts': [my_layout2], 'childWidgets': [widget4, widget5], 'items': [layout3]},
        }

    :param layout_data: Dictionary containing data used to initialize QLayouts.
    """
    for layout, data in layout_data.items():

        for key, val in data.items():
            match key:
                case 'childLayouts':
                    for _child in val:
                        layout.addChildLayout(_child)  # type: ignore

                case 'childWidgets':
                    for _child in val:
                        layout.addChildWidget(_child)  # type: ignore

                case 'items':
                    for _args in val:
                        if not isinstance(_args, Sequence):
                            _args = (_args,)

                        # Find the item's type and associated method.
                        item_type = type(_args[0])
                        for clazz, method_name in _layout_type_map.items():

                            if issubclass(item_type, clazz):
                                # Run method if exists and go to next item.
                                if hasattr(layout, method_name):
                                    getattr(layout, method_name)(*_args)
                                break


# noinspection PyUnresolvedReferences
def init_objects(object_data: dict[object, dict[str, Any]]) -> None:
    """Initialize :py:class:`QObject` attributes with the given data.

    object_data should be a dictionary structured like this::

        {
            widget1: {'text': 'some.translation.key', 'clicked': on_click},
            widget2: {'text': 'some.translation.key', 'size': {'fixed': (None, 400)}},
            widget3: {'text': string_value, 'pasted': on_paste, 'returnPressed': on_return},
            widget4: {'activated': when_activated, 'size': widget_size, 'items': (
                f'The Number "{i}"' for i in range(1, 11)
            )}
        }

    :param object_data: Dictionary containing data used to initialize basic QObject values.
    """
    # Initialize widget attributes
    for obj, data in object_data.items():
        special_keys = ('items', 'size')
        items, size_dict = (data.get(key) for key in special_keys)

        # Find setter method for all non specially-handled keys
        for key, val in data.items():
            if key in special_keys:
                continue  # Skip special keys

            if hasattr(obj, key):
                # Check if key is a signal on widget
                # If so, connect it to the given function
                if isinstance((attribute := getattr(obj, key)), SignalInstance):
                    if isinstance(val, Iterable):
                        for slot in val:
                            attribute.connect(slot)
                    else:
                        attribute.connect(val)
                    continue

            # Else call setter to update value
            # Capitalize first character of key
            setter_name: str = f'set{key[0].upper()}{key[1:]}'
            getattr(obj, setter_name)(val)

        if items is not None:
            if not isinstance(obj, QComboBox):
                # Set directly for non-dropdowns
                obj.setItems(items)  # type: ignore
            elif hasattr(obj, 'addItems'):
                obj.addItems(items)
            else:
                for key in items:
                    obj.addItem(key)

        # Set size
        if size_dict is not None:
            for size_type in ('minimum', 'maximum', 'fixed'):
                if size_dict.get(size_type) is not None:
                    size: QSize | Sequence[int] = size_dict.get(size_type)
                    if isinstance(size, QSize):
                        # For PySide6.QtCore.QSize objects
                        getattr(obj, f'set{size_type.title()}Size')(size)
                    elif isinstance(size, Sequence):
                        # For lists, tuples, etc. Set width and height separately.
                        # None can be used rather than int to skip a dimension.
                        if size[0]:
                            getattr(obj, f'set{size_type.title()}Width')(size[0])
                        if size[1]:
                            getattr(obj, f'set{size_type.title()}Height')(size[1])


def scroll_to_top(widget: QTextEdit) -> None:
    """Move text cursor to top of text editor."""
    cursor = widget.textCursor()
    cursor.setPosition(0)
    widget.setTextCursor(cursor)


def set_or_swap_icon(mapping: dict[str, QIcon], key: str, value: QIcon):
    """Given a mapping, replace a :py:class:`QIcon` value mapped to the given ``key`` with data from another.

    This keeps the same object references.
    """
    if key in mapping:
        mapping[key].swap(value)
    else:
        mapping[key] = value



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""``packaging`` package stripped out and modified for version.py."""

__all__ = (
    'BaseVersion',
    'InvalidVersion',
    'LegacyVersion',
    'parse_version',
    'Version',
    'VERSION_PATTERN',
)

from .version import BaseVersion
from .version import InvalidVersion
from .version import LegacyVersion
from .version import parse as parse_version
from .version import Version
from .version import VERSION_PATTERN



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Neutral namespace for constants. Meant to be star imported."""
from __future__ import annotations

__all__ = (
    'BYTE_UNITS',
    'HI_CACHE_PATH',
    'HI_CONFIG_PATH',
    'HI_PACKAGE_NAME',
    'HI_PATH_PATTERN',
    'HI_RESOURCE_PATH',
    'HI_SAMPLE_RESOURCE',
    'HI_TOKEN_PATH',
    'HI_URL_PATTERN',
    'HI_WEB_DUMP_PATH',
    'HI_WPAUTH_PATH',
    'MARKDOWN_IMG_LINK_PATTERN',
    'MARKDOWN_REF_LINK_PATTERN',
    'RFC_5646_PATTERN',
    'SUPPORTED_IMAGE_EXTENSIONS',
    'SUPPORTED_IMAGE_MIME_TYPES',
)

import re
from pathlib import Path
from typing import Final

# Mappings

BYTE_UNITS = {'Bytes': 1024**0, 'KiB': 1024**1, 'MiB': 1024**2, 'GiB': 1024**3, 'TiB': 1024**4}
"""Mapping of byte units to their respective powers of 1024."""

# Sets

SUPPORTED_IMAGE_EXTENSIONS: Final[frozenset[str]] = frozenset({
    'bmp', 'cur', 'gif', 'icns', 'ico', 'jpeg', 'jpg', 'pbm', 'pgm', 'png',
    'ppm', 'svg', 'svgz', 'tga', 'tif', 'tiff', 'wbmp', 'webp', 'xbm', 'xpm'
})
"""Set containing all image file extensions supported by application."""

SUPPORTED_IMAGE_MIME_TYPES: Final[frozenset[str]] = frozenset({
    'image/bmp', 'image/gif', 'image/jpeg', 'image/png', 'image/svg+xml', 'image/svg+xml-compressed',
    'image/tiff', 'image/vnd.microsoft.icon', 'image/vnd.wap.wbmp', 'image/webp', 'image/x-icns',
    'image/x-portable-bitmap', 'image/x-portable-graymap', 'image/x-portable-pixmap', 'image/x-tga',
    'image/x-xbitmap', 'image/x-xpixmap'
})
"""Set containing all image mime types supported by application."""

# Strings

HI_PACKAGE_NAME: Final[str] = __package__.split('.', maxsplit=1)[0]
"""The base package name for this application, for use in sub-packages."""

HI_SAMPLE_RESOURCE: Final[str] = 'Progression/file/Calendars/Seasons/SeasonCalendar.json'
"""Example resource. Is pre-filled in the search bar."""

# Paths

HI_CACHE_PATH: Final[Path] = Path.home() / '.cache/hi_getter'
"""Directory containing cached API results."""

HI_CONFIG_PATH: Final[Path] = Path.home() / '.config/hi_getter'
"""Directory containing user configuration data."""

HI_RESOURCE_PATH: Final[Path] = Path(__file__).parent / 'resources'
"""Directory containing application resources."""

HI_TOKEN_PATH: Final[Path] = HI_CONFIG_PATH / '.token'
"""File containing api token."""

HI_WEB_DUMP_PATH: Final[Path] = HI_CACHE_PATH / 'cached_requests'
"""Directory to dump web requests to."""

HI_WPAUTH_PATH: Final[Path] = HI_CONFIG_PATH / '.wpauth'
"""File containing auth key (Used for API token generation)."""

# Patterns

HI_PATH_PATTERN: Final[re.Pattern] = re.compile(
    r'[/\\]?(?P<pre_path>[\w\-_.]+[/\\]file[/\\])?'
    r'(?P<dir_name>(?:[\w\-_.]+[/\\])+)'
    r'(?P<file_name>[\w\-_]*\.\w+)')
"""Regex pattern for finding a resource path.
Finds quoted substrings with at least one folder name and file name (with a file extension).
"""

HI_URL_PATTERN: Final[re.Pattern] = re.compile(
    r'(?i)\b((?:https?:(?:/{1,3}|[a-z0-9%])|'
    r'[a-z0-9.\-]+[.][a-z]{2,}/)(?:[^\s()<>{}\[\]]+|'
    r'\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|'
    r'\(\S+?\))+(?:\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|'
    r'\(\S+?\)|[^\s`!()\[\]{};:\'".,<>?«»“”‘’])|'
    r'(?<!@)[a-z0-9]+(?:[.\-][a-z0-9]+)*[.][a-z]{2,}\b/?(?!@))')
"""Regex pattern for finding URLs.
Derived from https://gist.github.com/gruber/8891611.
"""

MARKDOWN_IMG_LINK_PATTERN: Final[re.Pattern] = re.compile(
    r'\[!\s*(?P<alt>\[[^\t\n\r()\[\]]*])\s*'
    r'(?P<image>\([\S ]+\))]\s*'
    r'(?P<target>[\[(][^\t\n\r()\[\]]*[)\]])')
"""Regex pattern for finding image links."""

MARKDOWN_REF_LINK_PATTERN: Final[re.Pattern] = re.compile(
    r'\[(?P<label>[^\[\]]+)] *: *'
    r'(?P<url>(?:(?:[a-zA-Z]+)?://)?'
    r'\w+(?:\.\w+)*(?::\d{1,5})?'
    r'(?:/[^\s()]*)?(?:\?(?:\w+=\w+&?)+)?)'
    r'(?: \"(?P<description>[^\"\t\n\r]*)\")?')
"""Regex pattern for finding markdown labels."""

RFC_5646_PATTERN: Final[re.Pattern] = re.compile(
    r'^(?=[a-z])(?:(?P<language>(?P<primary>[a-z]{2,3})(?:-(?P<extlang>[a-z]{3}))?)?'
    r'(?:-(?P<script>[a-z]{4}))?'
    r'(?:-(?P<region>[a-z]{2}|\d{3}))?'
    r'(?:-(?P<variants>(?:(?<![a-z\d])(?:[a-z\d]{5,8}|\d[a-z\d]{3})-?)+(?<!-)))?'
    r'(?:-(?P<extensions>(?:(?<![a-z\d])[a-wy-z\d]-[a-z\d]{2,8}-?)+(?<!-)))?)?'
    r'(?:-?(?<![a-z\d])(?P<private>x-(?:(?<![a-z\d])[a-z\d]{1,8}-?)+(?<!-)))?$',
    flags=re.IGNORECASE | re.ASCII)
"""Regex pattern for validating a language tag (langtag). Follows https://datatracker.ietf.org/doc/html/rfc5646

Uniqueness of variant subtags and extension singletons must be done outside of this pattern.

The language subtag is the primary subtag with the optional extlang subtag appended.

Most langtags are formatted as {language}-{REGION}.
Example langtags that follow RFC 5646::

    - "en" represents English ('en') with no region identifier.
    - "es-ES" represents Spanish ('es') as used in Spain ('ES').
    - "de-AT" represents German ('de') as used in Austria ('AT').
    - "sr-Latn" represents Serbian written using the Latin script.
    - "sr-Latn-RS" represents Serbian ('sr') written using Latin script ('Latn') as used in Serbia ('RS').
    - "es-419" represents Spanish ('es') appropriate to the UN-defined Latin America and Caribbean region ('419').
    - "sl-nedis" represents the Natisone or Nadiza dialect of Slovenian.
    - "de-CH-1996" represents German as used in Switzerland and as written using the spelling
            reform beginning in the year 1996 C.E.
    - en-a-bbb-x-a-ccc
    - x-private-tag-example
"""



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Module used for Event subscription."""
from __future__ import annotations

__all__ = (
    'Event',
    'EventBus',
)

from collections import defaultdict
from collections.abc import Callable
from typing import Generic
from typing import overload
from typing import TypeAlias
from typing import TypeVar

from .utils import get_parent_doc


class Event:
    """Normal event with no special abilities.

    You can also use the >> operator to fire an :py:class:`Event` on an :py:class:`EventBus`. Ex::

        Event() >> EventBus['foo']
    """

    __slots__: tuple[str, ...] = ()

    def __repr__(self) -> str:
        """Representation of the :py:class:`Event` with its attributes' values."""
        values = {attr: getattr(self, attr) for attr in self.__slots__}
        return f'<"{self.name}" Event {values=}>' if self.__slots__ else f'<Empty {self.name}>'

    def __rshift__(self, __bus: EventBus, /) -> None:
        """Syntax sugar for __bus.fire(event)."""
        return __bus.fire(event=self)

    def __rlshift__(self, __bus: EventBus, /) -> None:
        """Right shift if the EventBus.__lshift__ dunder is not working."""
        return self >> __bus

    @property
    def name(self) -> str:
        """Name of event.

        Defaults to class name.
        """
        return type(self).__name__

    @property
    def description(self) -> str:
        """Short description of the event.

        Defaults to the first line of the nearest type doc in the mro.
        """
        doc: str | None = self.__doc__
        # pylint: disable=unidiomatic-typecheck
        if doc is None and isinstance(self, Event) and type(self) is not Event:
            doc = get_parent_doc(type(self))
        return doc.splitlines()[0] if doc is not None else ''


_ET = TypeVar('_ET', bound=Event)  # Bound to Event. Can use Event subclass instances in place of Event instances.
_EventPredicate: TypeAlias = Callable[[_ET], bool]
_EventRunnable: TypeAlias = Callable[[_ET], None]


class _Subscribers(defaultdict[type[Event], list[tuple[
    _EventRunnable,          # Callable to run
    _EventPredicate | None   # Optional predicate to run callable
]]], Generic[_ET]):
    """Class which holds the event subscribers for an :py:class:`EventBus`."""

    def __init__(self) -> None:
        super().__init__(list)

    def __repr__(self) -> str:
        """Amount of subscribers for every event, encased in parentheses."""
        repr_: str = ''
        for event in self:
            repr_ += f'{event.__name__}[{len(self[event])}], '
        return f'({repr_.rstrip(", ")})'

    def add(self,
            event: type[_ET],
            callable_pair:
            tuple[_EventRunnable, _EventPredicate | None]
            ) -> None:
        """Add a callable pair to an Event subscriber list.

        :raises TypeError:
            If event is not a subclass of Event.
            If subscriber is not callable.
            If subscriber's predicate is defined but not a callable.
        """
        if not issubclass(event, Event):
            raise TypeError(f'event is not subclass to {Event}.')
        if not callable(callable_pair[0]):
            raise TypeError('subscriber is not callable.')
        if callable_pair[1] is not None and not callable(callable_pair[1]):
            raise TypeError('subscriber predicate is defined but not callable.')

        self[event].append(callable_pair)


class _EventBusMeta(type):
    """Metaclass for :py:class:`EventBus` type.

    Maps :py:class:`EventBus` objects to :py:class:`str` ids
    and allows accessing those ids using subscripts on :py:class:`EventBus`.
    """

    _id_bus_map: dict[str, EventBus] = {}

    @overload
    def __getitem__(cls, id: type[Event]) -> type[EventBus]:
        # FOR TYPE HINTING, IGNORE
        ...

    @overload
    def __getitem__(cls, id: str) -> EventBus:
        ...

    # pylint: disable=comparison-of-constants, compare-to-zero, no-value-for-parameter
    def __getitem__(cls, id) -> EventBus | type[EventBus]:
        # FOR TYPE HINTING, IGNORE
        if True is False:
            if issubclass(id, Event):
                return cls  # type: ignore

        if (bus := cls.get_bus(id)) is None:
            raise KeyError(f'{cls.__name__} "{id}" does not exist.')

        return bus

    def __setitem__(cls, id: str, bus: EventBus) -> None:
        """Set an :py:class:`EventBus` from the bus map.

        :raises TypeError:
            If id is not a str.
            If bus is not an Eventbus.
        """
        if not isinstance(id, str):
            raise TypeError(f'parameter {id=} is not of type {str}.')
        if not isinstance(bus, EventBus):
            raise TypeError(f'parameter {bus=} is not of type {EventBus}.')

        if bus.id is None:
            bus.id = id

        cls._id_bus_map[id.lower()] = bus

    def __delitem__(cls, id: str) -> None:
        """Delete an :py:class:`EventBus` from the bus map."""
        del cls._id_bus_map[id.lower()]

    def get_bus(cls, id: str, default: EventBus | None = None) -> EventBus | None:
        """Get bus from class map using the given id, with an optional default value.

        :param id: Case-insensitive id to search with.
        :param default: Default value if id resolves to None.
        :return: EventBus mapped to given id.
        """
        return cls._id_bus_map.get(id.lower(), default)


class EventBus(Generic[_ET], metaclass=_EventBusMeta):
    """Object that keeps track of all :py:class:`Callable` subscriptions to :py:class:`Event`'s.

    An Event Bus is an event-driven structure that stores both events and functions. When an
    event is fired, all functions "subscribed" to the event or any of its parent events are called with
    the event as a parameter.

    All :py:class:`EventBus`'s are stored in the class with a unique id.
    You can access created :py:class:`EventBus`'s with subscripts. ex::

        [Python Interactive Prompt]
            EventBus['foo'] -> None
            EventBus('foo') -> EventBus object at 0x000000001
            EventBus['foo'] -> EventBus object at 0x000000001
            del EventBus['foo']
            EventBus['foo'] -> None

    You can also use the << operator to fire an event, or an Event's >> operator. ex::

        [Python Interactive Prompt]
            # These fire an event
            EventBus['foo'] << Event()
            Event() >> EventBus['foo']

            # This raises an error:
            Event() << EventBus['foo']
    """

    def __init__(self, __id: str | None = None, /) -> None:
        """Create a new :py:class:`EventBus` object with a unique id.

        All ids are transformed to lowercase in the :py:class:`_EventBusMeta` id map.

        :param __id: id to register this instance as.
        :raises KeyError: When a bus with the given id already exists.
        """
        if __id is not None and type(self).get_bus(__id) is not None:
            raise KeyError(f'EventBus id "{__id}" is already registered in {type(type(self)).__name__}')

        self.id: str | None = __id
        self._subscribers: _Subscribers = _Subscribers()

        if __id is not None:
            type(self)[__id] = self

    def __lshift__(self, __event: _ET | type[_ET], /) -> None:
        """Syntax sugar for self.fire(event)."""
        return self.fire(__event)

    def __repr__(self) -> str:
        """Representation of the :py:class:`EventBus` with its id and :py:class:`_Subscribers`."""
        return f'<{type(self).__name__} id={self.id!r}; Subscribers={self._subscribers!r}>'

    def clear(self, event: type[_ET] | None = None) -> None:
        """Clear event _subscribers of a given type.

        None is treated as a wildcard and deletes ALL event _subscribers.

        :param event: Event type to clear.
        """
        if event is None:
            self._subscribers.clear()
        else:
            self._subscribers.pop(event)

    def fire(self, event: _ET | type[_ET]) -> None:
        """Fire all :py:class:`Callables` subscribed to the :py:class:`Event`'s :py:class:`type`.

        :py:class:`Event` subclasses call their parent's callables as well.
        Ex: ChildEvent will fire ParentEvent, but ParentEvent will not fire ChildEvent.

        If event is a :py:class:`type`, it is instantiated with no arguments.

        :param event: Event object passed to callables as an argument.
        """
        # Transform Event types to their default instances
        if isinstance(event, type) and issubclass(event, Event):
            event = event()

        # Run all current and parent event callables
        for event_type in self._subscribers:
            if isinstance(event, event_type):
                for e_callable_pair in self._subscribers[event_type]:
                    # Check predicate if one is given
                    if e_callable_pair[1] is None or e_callable_pair[1](event):

                        # Finally, call
                        e_callable_pair[0](event)

    # pylint: disable=useless-param-doc
    def subscribe(self,
                  __callable: _EventRunnable, /,
                  event: type[_ET],
                  event_predicate: _EventPredicate | None = None
                  ) -> None:
        """Subscribe a :py:class:`Callable` to an :py:class:`Event` type.

        By default, every time an :py:class:`Event` is fired, it will call the callable with the event as an argument.

        :param __callable: Callable to run.
        :param event: Event to subscribe to.
        :param event_predicate: Predicate to validate before running callable.
        :raises TypeError: If the given arguments are not valid.
        """
        callable_pair = (__callable, event_predicate)
        self._subscribers.add(event, callable_pair)



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""ScanSelectorDialog implementation."""
from __future__ import annotations

__all__ = (
    'ScanSelectorDialog',
)

import json
from pathlib import Path
from typing import Any

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...constants import *
from ...utils import init_layouts
from ...utils import init_objects
from ..aliases import app
from ..aliases import tr

_SCAN_FILTER: str = ';;'.join((
    'JSON Files (*.json)',
    'All files (*.*)'
))


class ScanSelectorDialog(QWidget):
    """A :py:class:`ScanSelectorDialog` which allows you to select JSON files to scan."""

    def __init__(self) -> None:
        """Create a new :py:class:`SignInDialog`."""
        super().__init__(None)

        self._init_ui()

    def _init_ui(self) -> None:
        self.filepaths_input = QLineEdit(self)
        self.scan_button = QPushButton(self)
        self.do_recursive = QCheckBox(self)

        init_objects({
            self: {
                'size': {'fixed': (500, 100)},
                'windowFlags': Qt.WindowType.Dialog,
                'windowModality': Qt.WindowModality.ApplicationModal,
                'windowIcon': self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogContentsView),
            },

            self.filepaths_input: {
                'maxLength': (2 ** 31) - 1,  # Max int for shiboken
            },

            (filepaths_selector := QPushButton(self)): {
                'clicked': self.select_files
            },

            self.do_recursive: {
                'checked': True
            },

            self.scan_button: {
                'clicked': self.scan_files,
            },

            (cancel_button := QPushButton(self)): {
                'clicked': self.close,
            }
        })

        app().init_translations({
            self.setWindowTitle: 'gui.scan_selector_dialog.title',
            self.filepaths_input.setPlaceholderText: 'gui.scan_selector_dialog.filepaths_placeholder',
            filepaths_selector.setText: 'gui.scan_selector_dialog.select_files',
            self.do_recursive.setText: 'gui.scan_selector_dialog.do_recursive',

            self.scan_button.setText: 'gui.scan_selector_dialog.scan',
            cancel_button.setText: 'gui.scan_selector_dialog.cancel'
        })

        init_layouts({
            (input_layout := QHBoxLayout()): {
                'items': (self.filepaths_input, filepaths_selector)
            },

            (bottom := QHBoxLayout()): {
                'items': (self.scan_button, cancel_button)
            },

            # Main layout
            QVBoxLayout(self): {
                'items': (input_layout, self.do_recursive, bottom)
            }
        })

    def _reset_widgets(self) -> None:
        self.filepaths_input.clear()

    # noinspection PyProtectedMember
    # pylint: disable=protected-access
    def scan_files(self) -> None:
        """Parse the input for paths, then read json data and tell the application's Client to scan it.

        :raises Exception: If an exception occurs when searching files, it is handled then reraised.
        """
        # If no filenames are selected, do nothing.
        if not (user_input := self.filepaths_input.text()):
            return

        app().warn_for('warnings.photosensitivity.scan')
        paths: tuple[Path] = tuple(Path(name.strip(' \'\",')) for name in user_input.split(', '))

        # Disconnect _on_finished_search so searched_paths isn't cleared prematurely
        app().client.finishedSearch.disconnect(app().client._on_finished_search)
        try:
            for i, path in enumerate(paths):
                path = path.expanduser().resolve(strict=True)
                data: dict[str, Any] = json.loads(path.read_bytes())

                if i == len(paths) - 1:
                    app().client.finishedSearch.connect(app().client._on_finished_search)

                app().client.start_handle_json(data, self.do_recursive.isChecked())

        # If an exception occurs, reconnect the _on_finished_search method and reraise
        except Exception as e:
            app().client.finishedSearch.connect(app().client._on_finished_search)
            raise e

        self.close()

    def select_files(self) -> None:
        """Select JSON files to scan through a file dialog."""
        names: list[str] = QFileDialog.getOpenFileNames(self, caption=tr('gui.scan_selector_dialog.select_files'),
                                                        dir=str(HI_WEB_DUMP_PATH),
                                                        filter=_SCAN_FILTER)[0]
        # If no filenames are selected, do nothing.
        if not names:
            return

        # Override input text with a string representation of selected files
        self.filepaths_input.setText(', '.join(f'"{Path(file)}"' for file in names))

    def showEvent(self, event: QShowEvent) -> None:
        """Reset widgets when dialog is shown to user."""
        super().showEvent(event)
        self._reset_widgets()



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.
#
# https://github.com/pypa/packaging/blob/main/LICENSE
"""This file has been refurbished by Cubicpath@Github to follow modern typing and documentation practices."""
from __future__ import annotations

__all__ = (
    'Infinity',
    'InfinityType',
    'NegativeInfinity',
    'NegativeInfinityType',
)

from typing import Any


class InfinityType:
    def __repr__(self) -> str:
        return 'Infinity'

    def __hash__(self) -> int:
        return hash(repr(self))

    def __lt__(self, other: Any) -> bool:
        return False

    def __le__(self, other: Any) -> bool:
        return False

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__)

    def __ne__(self, other: Any) -> bool:
        return not isinstance(other, self.__class__)

    def __gt__(self, other: Any) -> bool:
        return True

    def __ge__(self, other: Any) -> bool:
        return True

    def __neg__(self) -> NegativeInfinityType:
        return NegativeInfinity


Infinity = InfinityType()


class NegativeInfinityType:
    def __repr__(self) -> str:
        return '-Infinity'

    def __hash__(self) -> int:
        return hash(repr(self))

    def __lt__(self, other: Any) -> bool:
        return True

    def __le__(self, other: Any) -> bool:
        return True

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__)

    def __ne__(self, other: Any) -> bool:
        return not isinstance(other, self.__class__)

    def __gt__(self, other: Any) -> bool:
        return False

    def __ge__(self, other: Any) -> bool:
        return False

    def __neg__(self) -> InfinityType:
        return Infinity


NegativeInfinity = NegativeInfinityType()



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Main application window implementation."""
from __future__ import annotations

__all__ = (
    'AppWindow',
    'size_label_for',
)

import json
import string
import sys
from collections.abc import Callable
from pathlib import Path
from typing import Any

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...constants import *
from ...events import EventBus
from ...exception_hook import ExceptionEvent
from ...models import DeferredCallable
from ...models import DistributedCallable
from ...models import Singleton
from ...tomlfile import TomlEvents
from ...utils import init_layouts
from ...utils import init_objects
from ...utils import scroll_to_top
from ...utils import decode_url
from ...utils import http_code_map
from ..aliases import app
from ..aliases import tr
from ..menus import FileContextMenu
from ..menus import HelpContextMenu
from ..menus import ToolsContextMenu
from ..widgets import CacheExplorer
from ..widgets import HistoryComboBox
from ..widgets import ExceptionLogger
from ..widgets import ExternalTextBrowser
from ..widgets import PasteLineEdit
from .exception_reporter import ExceptionReporter


def size_label_for(num: int) -> str:
    """Return the best display unit to describe the given data's size.

    Ex: Bytes, KiB, MiB, GiB, TiB
    """
    display_unit = 'Bytes'
    for size_label, size in BYTE_UNITS.items():
        if num >= (size // 2):
            display_unit = size_label
        else:
            break
    return display_unit


class AppWindow(Singleton, QMainWindow):
    """Main window for the HaloInfiniteGetter application."""

    _singleton_base_type = QMainWindow
    _singleton_check_ref = False
    shown_key_warning: bool = False

    def __init__(self, size: QSize) -> None:
        """Create the window for the application."""
        super().__init__()
        app().client.receivedError.connect(self.update_error)
        app().client.receivedData.connect(self.update_image)
        app().client.receivedJson.connect(self.update_text)

        self.current_image: QPixmap | None = None
        self.detached: dict[str, QMainWindow | None] = {'media': None, 'text': None}
        self.resize(size)

        EventBus['settings'].subscribe(
            DeferredCallable(self.resize_image),
            TomlEvents.Set, lambda event: event.key.startswith('gui/media_output/'))

        EventBus['settings'].subscribe(
            lambda event: self.text_output.setLineWrapMode(event.new),
            TomlEvents.Set, lambda event: event.key == 'gui/text_output/line_wrap_mode')

        self._init_toolbar()
        self._init_ui()

    def _init_toolbar(self) -> None:
        """Initialize toolbar widgets."""
        from .settings import SettingsWindow

        def context_menu_handler(menu_class: type[QMenu]) -> None:
            """Create a new :py:class:`QMenu` and show it at the cursor's position.

            :raises TypeError: If menu_class is not an instance of QMenu.
            """
            if not issubclass(menu_class, QMenu):
                raise TypeError(f'{menu_class} is not a subclass of {QMenu}')

            menu: QMenu = menu_class(self)
            menu.exec(self.cursor().pos())
            menu.deleteLater()

        init_objects({
            (menu_bar := QToolBar(self)): {},

            (status_bar := QToolBar(self)): {
                'movable': False,
            },

            (file := QAction(self)): {
                'menuRole': QAction.MenuRole.ApplicationSpecificRole,
                'triggered': DeferredCallable(context_menu_handler, FileContextMenu)
            },

            (settings := QAction(self)): {
                'menuRole': QAction.MenuRole.PreferencesRole,
                'triggered': DistributedCallable((
                    SettingsWindow.instance().show,            # pyright: ignore[reportGeneralTypeIssues]
                    SettingsWindow.instance().activateWindow,  # pyright: ignore[reportGeneralTypeIssues]
                    SettingsWindow.instance().raise_           # pyright: ignore[reportGeneralTypeIssues]
                ))
            },

            (tools := QAction(self)): {
                'menuRole': QAction.MenuRole.ApplicationSpecificRole,
                'triggered': DeferredCallable(context_menu_handler, ToolsContextMenu)
            },

            (help := QAction(self)): {
                'menuRole': QAction.MenuRole.AboutRole,
                'triggered': DeferredCallable(context_menu_handler, HelpContextMenu)
            },

            (logger := ExceptionLogger(self)): {
                'size': {'fixed': (None, 20)},
                'clicked': DistributedCallable((
                    logger.reporter.show,
                    logger.reporter.activateWindow,
                    logger.reporter.raise_
                ))
            }
        })

        app().init_translations({
            menu_bar.setWindowTitle: 'gui.menu_bar.title',
            status_bar.setWindowTitle: 'gui.status_bar.title',
            file.setText: 'gui.menus.file',
            settings.setText: 'gui.menus.settings',
            tools.setText: 'gui.menus.tools',
            help.setText: 'gui.menus.help',
            logger.label.setText: 'gui.status.default'
        })

        self.exception_reporter: ExceptionReporter = logger.reporter

        self.addToolBar(Qt.ToolBarArea.TopToolBarArea, menu_bar)
        self.addToolBar(Qt.ToolBarArea.BottomToolBarArea, status_bar)

        EventBus['exceptions'].subscribe(lambda e: logger.label.setText(f'{e.exception}...'), ExceptionEvent)
        for action in (file, settings, tools, help):
            menu_bar.addSeparator()
            menu_bar.addAction(action)

        status_bar.addWidget(logger)
        status_bar.addSeparator()
        status_bar.addWidget(logger.label)

    # noinspection PyTypeChecker
    def _init_ui(self) -> None:
        """Initialize the UI, including Layouts and widgets."""

        def setup_detached_window(id_: str, frame: QFrame, handler: Callable, title: str | None = None) -> QMainWindow:
            """Set up a detached window, with the layout represented as a :py:class:`QFrame`.

            :param id_: unique name for the window
            :param frame: Set the window's central widget as this QFrame.
            :param handler: Callable to execute when closed, to reattach the frame to the parent window.
            :param title: The window title.
            """
            init_objects({
                (window := QMainWindow()): {
                    'windowTitle': title if title is not None else self.windowTitle(),
                    'centralWidget': frame,
                    'size': {'minimum': (300, 200)}
                }
            })

            def close_event(*_) -> None:
                if self.detached[id_]:
                    handler()

            window.closeEvent = close_event  # pyright: ignore[reportGeneralTypeIssues]
            return window

        def toggle_media_detach() -> None:
            """Handle the detaching and reattaching the media output."""
            if self.detached['media'] is None:
                self.detached['media'] = window = setup_detached_window(
                    'media',
                    self.media_frame, toggle_media_detach,
                    tr('gui.outputs.image.detached')
                )
                self.image_detach_button.setText(tr('gui.outputs.reattach'))
                window.resizeEvent = DeferredCallable(self.resize_image)
                window.show()
            else:
                window = self.detached['media']
                self.detached['media'] = None
                window.close()

                self.outputs.insertWidget(0, self.media_frame)
                self.image_detach_button.setText(tr('gui.outputs.detach'))

        def toggle_text_detach() -> None:
            """Handle the detaching and reattaching the text output."""
            if self.detached['text'] is None:
                self.detached['text'] = window = setup_detached_window(
                    'text',
                    self.text_frame, toggle_text_detach,
                    tr('gui.outputs.text.detached')
                )
                self.text_detach_button.setText(tr('gui.outputs.reattach'))
                window.show()
            else:
                window = self.detached['text']
                self.detached['text'] = None
                window.close()

                self.outputs.insertWidget(-1, self.text_frame)
                self.text_detach_button.setText(tr('gui.outputs.detach'))

        def clear_current_pixmap() -> None:
            """Clear the current image from the media output."""
            self.image_size_label.setText(tr('gui.outputs.image.label_empty'))
            self.clear_picture.setDisabled(True)
            self.copy_picture.setDisabled(True)
            self.media_output.scene().clear()
            self.current_image = None

        def clear_current_text() -> None:
            """Clear the current text from the text output."""
            self.text_size_label.setText(tr('gui.outputs.text.label_empty'))
            self.clear_text.setDisabled(True)
            self.copy_text.setDisabled(True)
            self.text_output.setDisabled(True)
            self.text_output.clear()

        def next_in_history() -> None:
            """View the next license."""
            # Qt automatically rolls over to -1, so we handle it like prev_in_history
            self.input_field.setCurrentIndex(self.input_field.currentIndex() + 1)
            if self.input_field.currentIndex() < 0:
                self.input_field.setCurrentIndex(0)
            self.use_input()

        def prev_in_history() -> None:
            """View the previous license."""
            self.input_field.setCurrentIndex(self.input_field.currentIndex() - 1)
            if self.input_field.currentIndex() < 0:
                self.input_field.setCurrentIndex(self.input_field.count() - 1)
            self.use_input()

        def open_file_in_view(file_path: str) -> None:
            if not Path(file_path).is_file():
                return

            if file_path.endswith(tuple(SUPPORTED_IMAGE_EXTENSIONS)):
                self.update_image(file_path, Path(file_path).read_bytes())
            else:
                self.update_text(file_path, Path(file_path).read_text(encoding='utf8'))

        # Define widget attributes
        # Cannot be defined in init_objects() as walrus operators are not allowed for object attribute assignment.
        # This works in the standard AST, but is a seemingly arbitrary limitation set by the interpreter.
        # See:
        # https://stackoverflow.com/questions/64055314/why-cant-pythons-walrus-operator-be-used-to-set-instance-attributes#answer-66617839
        (
            self.input_field, self.image_size_label, self.text_size_label,
            self.image_detach_button, self.text_detach_button, self.clear_picture, self.copy_picture,
            self.clear_text, self.copy_text, self.media_frame, self.text_frame,
            self.media_output, self.text_output, self.cache_explorer
        ) = (
            HistoryComboBox(self), QLabel(self), QLabel(self),
            QPushButton(self), QPushButton(self), QPushButton(self), QPushButton(self),
            QPushButton(self), QPushButton(self), QFrame(self), QFrame(self),
            QGraphicsView(self), ExternalTextBrowser(self), CacheExplorer(self)
        )

        init_objects({
            # Labels
            self.image_size_label: {
                'size': {'minimum': (50, None)}
            },
            self.text_size_label: {
                'size': {'minimum': (50, None)}
            },

            # Buttons
            self.image_detach_button: {
                'size': {'maximum': (80, None)},
                'clicked': toggle_media_detach
            },
            self.text_detach_button: {
                'size': {'maximum': (80, None)},
                'clicked': toggle_text_detach
            },
            self.clear_picture: {
                'disabled': True,
                'size': {'maximum': (80, None), 'minimum': (40, None)},
                'clicked': clear_current_pixmap
            },
            self.copy_picture: {
                'disabled': True,
                'size': {'maximum': (160, None), 'minimum': (80, None)},
                'clicked': DeferredCallable(app().clipboard().setPixmap, lambda: self.current_image)
            },
            self.clear_text: {
                'disabled': True,
                'size': {'maximum': (80, None), 'minimum': (40, None)},
                'clicked': clear_current_text
            },
            self.copy_text: {
                'disabled': True,
                'size': {'maximum': (160, None), 'minimum': (80, None)},
                'clicked': DeferredCallable(app().clipboard().setText, self.text_output.toPlainText)
            },
            (get_button := QPushButton(self)): {
                'size': {'maximum': (40, None)},
                'clicked': self.use_input
            },
            (scan_button := QPushButton(self)): {
                'size': {'maximum': (55, None)},
                'clicked': self.scan_input
            },

            # Line editors
            self.input_field: {
                'items': (HI_SAMPLE_RESOURCE,)
            },
            self.input_field.lineEdit(): {
                'returnPressed': self.use_input
            },
            (subdomain_field := PasteLineEdit(self)): {
                'text': app().client.sub_host, 'disabled': True,
                'size': {'fixed': (125, None)}
            },
            (root_folder_field := PasteLineEdit(self)): {
                'text': app().client.parent_path, 'disabled': True,
                'size': {'fixed': (28, None)}
            },

            self.cache_explorer: {
                'openFileInView': open_file_in_view
            },

            # Outputs
            self.media_output: {
                'disabled': True, 'size': {'minimum': (None, 28)},
                'scene': QGraphicsScene(self), 'autoFillBackground': False,
                'horizontalScrollBarPolicy': Qt.ScrollBarPolicy.ScrollBarAlwaysOff,
                'verticalScrollBarPolicy': Qt.ScrollBarPolicy.ScrollBarAlwaysOff
            },
            self.text_output: {
                'disabled': True, 'size': {'minimum': (None, 28)},
                'lineWrapMode': QTextEdit.LineWrapMode(app().settings['gui/text_output/line_wrap_mode']),
                'openLinks': False, 'anchorClicked': self.navigate_to
            }
        })

        app().init_translations({
            # Labels
            self.image_size_label.setText: 'gui.outputs.image.label_empty',
            self.text_size_label.setText: 'gui.outputs.text.label_empty',

            # Buttons
            self.image_detach_button.setText: 'gui.outputs.detach',
            self.text_detach_button.setText: 'gui.outputs.detach',
            self.clear_picture.setText: 'gui.outputs.clear',
            self.copy_picture.setText: 'gui.outputs.image.copy',
            self.clear_text.setText: 'gui.outputs.clear',
            self.copy_text.setText: 'gui.outputs.text.copy',
            get_button.setText: 'gui.input_field.get',
            scan_button.setText: 'gui.input_field.scan'
        })

        self.text_output.connect_key_to(Qt.Key.Key_Left, prev_in_history)
        self.text_output.connect_key_to(Qt.Key.Key_Right, next_in_history)
        self.outputs = QHBoxLayout()

        init_layouts({
            # Text widget layouts
            (text_bottom := QHBoxLayout()): {
                'items': [
                    (self.clear_text, Qt.AlignmentFlag.AlignLeft),
                    (self.copy_text, Qt.AlignmentFlag.AlignLeft)
                ]
            },
            (text_top := QHBoxLayout()): {
                'items': [
                    (self.text_size_label, Qt.AlignmentFlag.AlignLeft),
                    (self.text_detach_button, Qt.AlignmentFlag.AlignRight)
                ]
            },
            QVBoxLayout(self.text_frame): {
                'items': [
                    text_top, self.text_output, text_bottom
                ]
            },

            # Media widget layouts
            (media_bottom := QHBoxLayout()): {
                'items': [
                    (self.clear_picture, Qt.AlignmentFlag.AlignLeft),
                    (self.copy_picture, Qt.AlignmentFlag.AlignLeft)
                ]
            },
            (media_top := QHBoxLayout()): {
                'items': [
                    (self.image_size_label, Qt.AlignmentFlag.AlignLeft),
                    (self.image_detach_button, Qt.AlignmentFlag.AlignRight)
                ]
            },
            QVBoxLayout(self.media_frame): {
                'items': [
                    media_top, self.media_output, media_bottom
                ]
            },

            # Middle parent layouts
            self.outputs: {
                'items': [
                    self.media_frame, self.text_frame
                ]
            },
            (middle := QHBoxLayout()): {
                'items': [
                    (self.cache_explorer, Qt.AlignmentFlag.AlignLeft),
                    (self.outputs, Qt.AlignmentFlag.AlignHCenter)
                ]
            },

            # Top widget layout
            (top := QHBoxLayout()): {
                'items': [
                    subdomain_field, root_folder_field, self.input_field,
                    get_button, scan_button
                ]
            },

            # Main layout
            (layout := QGridLayout()): {
                'items': [
                    (top, 10, 0, Qt.AlignmentFlag.AlignTop),
                    (middle, 20, 0, Qt.AlignmentFlag.AlignHCenter),
                    # (bottom, 30, 0, Qt.AlignmentFlag.AlignBottom)
                ]
            }
        })

        init_objects({
            (main_widget := QWidget()): {'layout': layout},

            text_bottom: {'spacing': 5},
            top: {'spacing': 2},

            self: {'centralWidget': main_widget}
        })

    def navigate_to(self, path: QUrl) -> None:
        """Set input field text to path and get resource."""
        str_path = decode_url(path.toDisplayString())
        has_volume_id = str_path[0] in string.ascii_letters and str_path[1:].startswith(':\\')

        if (sys.platform == 'win32' and has_volume_id) or sys.platform.startswith('linux') and str_path.startswith('/'):
            # Open local absolute resource locations
            QDesktopServices.openUrl(QUrl(str_path))
        else:
            self.input_field.addItem(str_path)
            self.input_field.setCurrentIndex(self.input_field.count() - 1)
            self.use_input()

    def use_input(self, scan: bool = False) -> None:
        """Use the current input field's text to search through the Client for data.

        Automatically handles media and text data.

        :param scan: Whether to recursively scan a resource.
        """
        if not (user_input := self.input_field.currentText()):
            return

        if search_path := user_input.strip():
            if scan:
                app().client.recursive_search(search_path)
            else:
                app().client.get_hi_data(search_path)

    def update_image(self, _: str, data: bytes) -> None:
        """Update the image output with the given data."""
        display_unit: str = size_label_for(len(data))
        display_size: float = round(len(data) / BYTE_UNITS[display_unit], 4)

        self.clear_picture.setDisabled(False)
        self.copy_picture.setDisabled(False)
        self.current_image = QPixmap()
        self.current_image.loadFromData(data)
        self.image_size_label.setText(tr(
            'gui.outputs.image.label',
            self.current_image.size().width(), self.current_image.size().height(),  # Image dimensions
            display_size, display_unit                                              # Size in given unit
        ))
        self.resize_image()

    def update_text(self, search_path: str, data: str | dict[str, Any]) -> None:
        """Update the text output with the given data."""
        if isinstance(data, dict):
            data = json.dumps(data, indent=2)

        scroll_to_top(self.text_output)
        self.clear_text.setDisabled(False)
        self.copy_text.setDisabled(False)
        self.text_output.setDisabled(False)
        self.text_output.clear()

        display_unit: str = size_label_for(len(data.encode('utf8', errors='ignore')))
        display_size: float = round(len(data) / BYTE_UNITS[display_unit], 4)

        # Load up to 8 MiB of text data
        output = data if len(data) <= BYTE_UNITS['MiB'] * 8 else tr(
            'gui.outputs.text.errors.too_large',
            app().client.to_os_path(search_path)
        )

        replaced = set()
        for match in HI_PATH_PATTERN.finditer(output):
            if (match := match[0].replace('"', '')) not in replaced:
                output = output.replace(match, f'<a href="{match}" style="color: #2A5DB0">{match}</a>')
                replaced.add(match)

        self.text_size_label.setText(tr(
            'gui.outputs.text.label',
            len(data.splitlines()), len(data),  # Line and character count
            display_size, display_unit          # Size in given unit
        ))

        self.text_output.setHtml(
            '<body style="white-space: pre-wrap">'
            f'{output}'
            '</body>'
        )

    def update_error(self, search_path: str, data: int) -> None:
        """Update the text output with the given data."""
        scroll_to_top(self.text_output)
        self.clear_text.setDisabled(False)
        self.copy_text.setDisabled(False)
        self.text_output.setDisabled(False)
        self.text_output.clear()

        error: str = tr(
            'gui.outputs.text.errors.http',
            app().client.api_root + search_path,  # Search path
            data, http_code_map[data][0],         # Error code and phrase
            http_code_map[data][1]                # Error description
        )
        self.text_size_label.setText(tr('gui.outputs.text.label_empty'))
        self.text_output.setPlainText(error)

    def resize_image(self) -> None:
        """Refresh the media output with a resized version of the current image."""
        if self.current_image is not None:
            new = self.current_image.copy()
            self.media_output.scene().clear()  # Clear buffer, otherwise causes memory leak

            if self.current_image.size() != self.media_output.viewport().size():
                # Create a new image from the copied source image, scaled to fit the window.
                new = new.scaled(
                    self.media_output.viewport().size(),
                    Qt.AspectRatioMode(app().settings['gui/media_output/aspect_ratio_mode']),
                    Qt.TransformationMode(app().settings['gui/media_output/transformation_mode'])
                )

            # Add image to buffer
            self.media_output.scene().addPixmap(new)

    def scan_input(self) -> None:
        """Scan the search path.

        Displays a photosensitivity warning for first-time use.
        """
        app().warn_for('warnings.photosensitivity.scan')
        app().client.searched_paths.clear()
        self.use_input(scan=True)

    # # # # # Events

    def show(self) -> None:
        """After window is displayed, show warnings if not already warned."""
        super().show()

        if app().first_launch:
            app().windows['readme_viewer'].show()
            app().show_dialog('information.first_launch', self)

        elif not self.shown_key_warning and app().client.token is None:
            app().show_dialog('warnings.empty_token', self)
            self.__class__.shown_key_warning = True

    def resizeEvent(self, event: QResizeEvent) -> None:
        """Resize image on resize of window."""
        super().resizeEvent(event)
        self.resize_image()
        event.accept()

    def closeEvent(self, event: QCloseEvent) -> None:
        """Close all detached/children windows and quit application."""
        super().closeEvent(event)
        # Remember window size
        app().settings.reload()
        app().settings['gui/window/x_size'] = self.size().width()
        app().settings['gui/window/y_size'] = self.size().height()
        app().settings.save()

        app().quit()
        event.accept()



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""Main networking package for HaloInfiniteGetter."""

__all__ = (
    'Client',
    'gc_response',
    'KNOWN_HEADERS',
    'NetworkSession',
    'Request',
    'Response',
    'VersionChecker',
)

from .client import Client
from .manager import gc_response
from .manager import KNOWN_HEADERS
from .manager import NetworkSession
from .manager import Request
from .manager import Response
from .version_check import VersionChecker



###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""ReadmeViewer implementation."""
from __future__ import annotations

__all__ = (
    'ChangelogViewer',
)

import random

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ..._version import __version__
from ...models import DeferredCallable
from ...models import DistributedCallable
from ...network import Response
from ...utils import init_layouts
from ...utils import init_objects
from ..aliases import app
from ..aliases import tr
from ..widgets import ExternalTextBrowser


class ChangelogViewer(QWidget):
    """Widget that formats and shows the project's README.md, stored in the projects 'Description' metadata tag."""

    def __init__(self, *args, **kwargs) -> None:
        """Create a new :py:class:`ChangelogViewer` and initialize UI."""
        super().__init__(*args, **kwargs)
        self.resize(QSize(600, 800))
        self.setWindowTitle(tr('gui.changelog_viewer.title'))
        self.setWindowIcon(app().get_theme_icon('message_information') or
                           self.style().standardIcon(QStyle.StandardPixmap.SP_DialogApplyButton))

        self.changelog_url: str = 'https://raw.githubusercontent.com/Cubicpath/HaloInfiniteGetter/master/CHANGELOG.md'
        self.latest_version: str = __version__
        self._init_ui()

    def _init_ui(self) -> None:
        self.text_browser = ExternalTextBrowser(self)
        self.version_label = QLabel(self)

        init_objects({
            self.text_browser: {
                'font': QFont(self.text_browser.font().family(), 10),
                'openExternalLinks': True
            },

            (check_latest_button := QPushButton(self)): {
                'size': {'fixed': (20, 20)},
                'icon': self.style().standardIcon(QStyle.StandardPixmap.SP_BrowserReload),
                'clicked': DistributedCallable((
                    DeferredCallable(app().version_checker.check_version),
                    check_latest_button.setDisabled,
                    self.version_label.setDisabled
                ), True)
            },

            (spacer := QSpacerItem(0, 0, hData=QSizePolicy.Policy.MinimumExpanding)): {},

            (github_button := QPushButton(self)): {
                'size': {'fixed': (180, None)},
                'clicked': DeferredCallable(
                    QDesktopServices.openUrl,
                    QUrl('https://github.com/Cubicpath/HaloInfiniteGetter/blob/master/CHANGELOG.md')
                )
            },

            (releases_button := QPushButton(self)): {
                'size': {'fixed': (110, None)},
                'clicked': DeferredCallable(
                    QDesktopServices.openUrl,
                    QUrl('https://github.com/Cubicpath/HaloInfiniteGetter/releases')
                )
            }
        })

        app().init_translations({
            self.setWindowTitle: 'gui.changelog_viewer.title',
            self.version_label.setText: ('gui.changelog_viewer.current_version', self.latest_version, __version__),
            github_button.setText: 'gui.changelog_viewer.github',
            releases_button.setText: 'gui.changelog_viewer.releases'
        })

        init_layouts({
            (buttons := QHBoxLayout()): {
                'items': [check_latest_button, self.version_label, spacer, github_button, releases_button]
            },

            # Main layout
            QVBoxLayout(self): {
                'items': [buttons, self.text_browser]
            }
        })

        app().version_checker.newerVersion.connect(self.update_changelog)
        app().version_checker.newerVersion.connect(self.update_latest_version)
        app().version_checker.checked.connect(lambda: QTimer.singleShot(random.randint(250, 500), DistributedCallable((
            check_latest_button.setDisabled,
            self.version_label.setDisabled
        ), False)))

        self.update_changelog()

    def update_changelog(self) -> None:
        """Update the displayed text to the data from the changelog url."""

        def handle_reply(reply: Response):
            text: str = reply.text
            # Add separators between versions and remove primary header
            text = '## ' + '\n-----\n## '.join(text.split('\n## ')[1:])

            # The Markdown Renderer doesn't accept headers with ref_links, so remove them
            new_lines = []
            for line in text.splitlines():
                if line.startswith('## '):
                    line = line.replace('[', '', 1).replace(']', '', 1)

                new_lines.append(line)
            text = '\n'.join(new_lines)

            self.text_browser.set_hot_reloadable_text(text, 'markdown')

        app().session.get(self.changelog_url, finished=handle_reply)

    def update_latest_version(self, _, version: str) -> None:
        """Update the latest version displayed at the top of the changelog to the given string.

        :param version: Latest version string to display
        """
        self.latest_version = version

        text = tr('gui.changelog_viewer.current_version', self.latest_version, __version__)
        self.version_label.setText(f'<b>{text.split("|")[0]}</b>|{text.split("|")[1]}')
